from uuid import uuid4

from django.test import TestCase

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission

from pulpcore.app.models import Group, Remote, Repository
from pulpcore.app.models.role import Role
from pulpcore.app.role_util import assign_role, remove_role, get_objects_for_user


User = get_user_model()


class UserRoleTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create(username=uuid4())
        self.group = Group.objects.create(name=uuid4())
        self.group.user_set.add(self.user)
        self.role1 = Role.objects.create(name="role1")
        self.role1.permissions.add(
            Permission.objects.get(content_type__app_label="core", codename="view_repository")
        )
        self.role2 = Role.objects.create(name="role2")
        self.role2.permissions.add(
            Permission.objects.get(content_type__app_label="core", codename="view_remote")
        )
        self.repository = Repository.objects.create(name=uuid4())
        self.repository2 = Repository.objects.create(name=uuid4())
        self.remote = Remote.objects.create(name=uuid4())
        self.remote2 = Remote.objects.create(name=uuid4())

    def test_user_no_role(self):
        self.assertFalse(self.user.has_perm("core.view_repository"))
        self.assertFalse(self.user.has_perm("core.view_repository", self.repository))
        self.assertFalse(self.user.has_perm("core.view_remote"))
        self.assertFalse(self.user.has_perm("core.view_remote", self.remote))
        self.assertEqual(self.user.get_all_permissions(), set())
        self.assertEqual(self.user.get_all_permissions(self.repository), set())

    def test_user_object_role(self):
        assign_role("role1", self.user, self.repository)
        self.assertFalse(self.user.has_perm("core.view_repository"))
        self.assertTrue(self.user.has_perm("core.view_repository", self.repository))
        self.assertEqual(self.user.get_all_permissions(), set())
        self.assertEqual(self.user.get_all_permissions(self.repository), {"view_repository"})
        remove_role("role1", self.user, self.repository)

    def test_user_role(self):
        assign_role("role1", self.user)
        self.assertTrue(self.user.has_perm("core.view_repository"))
        self.assertFalse(self.user.has_perm("core.view_repository", self.repository))
        self.assertEqual(self.user.get_all_permissions(), {"core.view_repository"})
        self.assertEqual(self.user.get_all_permissions(self.repository), set())
        remove_role("role1", self.user)

    def test_group_object_role(self):
        assign_role("role2", self.group, self.remote)
        self.assertFalse(self.user.has_perm("core.view_remote"))
        self.assertTrue(self.user.has_perm("core.view_remote", self.remote))
        self.assertEqual(self.user.get_all_permissions(), set())
        self.assertEqual(self.user.get_all_permissions(self.remote), {"view_remote"})
        remove_role("role2", self.group, self.remote)

    def test_group_role(self):
        assign_role("role2", self.group)
        self.assertTrue(self.user.has_perm("core.view_remote"))
        self.assertFalse(self.user.has_perm("core.view_remote", self.remote))
        self.assertEqual(self.user.get_all_permissions(), {"core.view_remote"})
        self.assertEqual(self.user.get_all_permissions(self.remote), set())
        remove_role("role2", self.group)

    def test_combination_role(self):
        assign_role("role1", self.user, self.repository)
        assign_role("role2", self.group)
        self.assertEqual(self.user.get_all_permissions(), {"core.view_remote"})
        self.assertEqual(self.user.get_all_permissions(self.repository), {"view_repository"})
        self.assertEqual(self.user.get_all_permissions(self.remote), set())
        self.assertEqual(
            set(
                get_objects_for_user(
                    self.user, "core.view_repository", Repository.objects.all()
                ).values_list("pk", flat=True)
            ),
            {self.repository.pk},
        )
        self.assertEqual(
            set(
                get_objects_for_user(
                    self.user, "core.view_remote", Remote.objects.all()
                ).values_list("pk", flat=True)
            ),
            {self.remote.pk, self.remote2.pk},
        )
        remove_role("role2", self.group)



from import_export import fields
from import_export.widgets import ForeignKeyWidget
from logging import getLogger

from pulpcore.app.models.content import (
    Artifact,
    Content,
    ContentArtifact,
)
from pulpcore.app.models.repository import Repository
from pulpcore.constants import ALL_KNOWN_CONTENT_CHECKSUMS
from pulpcore.plugin.importexport import QueryModelResource


log = getLogger(__name__)


#
# Artifact and Repository are different from other import-export entities, in that they are not
# repo-version-specific.
#
class ArtifactResource(QueryModelResource):
    """Resource for import/export of artifacts."""

    def before_import_row(self, row, **kwargs):
        """
        Sets digests to None if they are blank strings.

        Args:
            row (tablib.Dataset row): incoming import-row representing a single Variant.
            kwargs: args passed along from the import() call.

        """
        # the export converts None to blank strings but sha384 and sha512 have unique constraints
        # that get triggered if they are blank. convert checksums back into None if they are blank.
        for checksum in ALL_KNOWN_CONTENT_CHECKSUMS:
            if row[checksum] == "":
                row[checksum] = None

    class Meta:
        model = Artifact
        exclude = (
            "pulp_id",
            "pulp_created",
            "pulp_last_updated",
        )
        import_id_fields = ("sha256",)


class RepositoryResource(QueryModelResource):
    class Meta:
        model = Repository
        import_id_fields = ("name",)
        exclude = (
            "pulp_id",
            "pulp_created",
            "pulp_last_updated",
            "content",
        )


class ContentArtifactResource(QueryModelResource):
    """
    Handles import/export of the ContentArtifact model.

    ContentArtifact is different from other import-export entities because it has no 'natural key'
    other than a pulp_id, which aren't shared across instances. We do some magic to link up
    ContentArtifacts to their matching (already-imported) Content.

    Some plugin-models have sub-repositories. We take advantage of the content-mapping
    machinery to account for those contentartifacts as well.
    """

    artifact = fields.Field(
        column_name="artifact", attribute="artifact", widget=ForeignKeyWidget(Artifact, "sha256")
    )

    def __init__(self, repo_version=None, content_mapping=None):
        self.content_mapping = content_mapping
        super().__init__(repo_version)

    def before_import_row(self, row, **kwargs):
        """
        Fixes the content-ptr of an incoming content-artifact row at import time.

        Finds the 'original uuid' of the Content for this row, looks it up as the
        'upstream_id' of imported Content, and then replaces the Content-pk with its
        (new) uuid.

        Args:
            row (tablib.Dataset row): incoming import-row representing a single ContentArtifact.
            kwargs: args passed along from the import() call.

        Returns:
            (tablib.Dataset row): row that now points to the new downstream uuid for its content.
        """

        linked_content = Content.objects.get(upstream_id=row["content"])
        row["content"] = str(linked_content.pulp_id)

    def set_up_queryset(self):
        vers_content = ContentArtifact.objects.filter(content__in=self.repo_version.content)
        if self.content_mapping:
            all_content = []
            for content_ids in self.content_mapping.values():
                all_content.extend(content_ids)
            vers_content = vers_content.union(
                ContentArtifact.objects.filter(content__in=all_content)
            )
        return vers_content.order_by("content", "relative_path")

    class Meta:
        model = ContentArtifact
        import_id_fields = (
            "content",
            "relative_path",
        )
        exclude = (
            "pulp_created",
            "pulp_last_updated",
            "_artifacts",
            "pulp_id",
        )



# Generated by Django 3.2.9 on 2021-12-08 11:33

from django.conf import settings
from django.db import migrations


def reapplabel_group_permissions_up(apps, schema_editor, up=True):
    User = apps.get_model(settings.AUTH_USER_MODEL)

    PERMISSION_CLASSES = [
        (User.user_permissions.through._meta.app_label, User.user_permissions.through._meta.model_name),
        ("auth", "Group_permissions"),
        ("guardian", "UserObjectPermission"),
        ("guardian", "GroupObjectPermission"),
    ]

    PERMISSION_NAMES = ["add", "change", "delete", "view"]

    Permission = apps.get_model("auth", "Permission")
    ContentType = apps.get_model("contenttypes", "ContentType")
    AuthGroup = apps.get_model("auth", "Group")
    CoreGroup = apps.get_model("core", "Group")
    auth_ctype = ContentType.objects.get_for_model(AuthGroup, for_concrete_model=False)
    core_ctype = ContentType.objects.get_for_model(CoreGroup, for_concrete_model=False)

    perm_classes = [apps.get_model(class_app, class_name) for class_app, class_name in PERMISSION_CLASSES]

    for perm_name in PERMISSION_NAMES:
        auth_perm, _ = Permission.objects.get_or_create(content_type=auth_ctype, codename=f"{perm_name}_group", defaults={"name": f"Can {perm_name} group"})
        core_perm, _ = Permission.objects.get_or_create(content_type=core_ctype, codename=f"{perm_name}_group", defaults={"name": f"Can {perm_name} group"})
        for perm_class in perm_classes:
            if up:
                perm_class.objects.filter(permission=auth_perm).update(permission=core_perm)
            else:
                perm_class.objects.filter(permission=core_perm).update(permission=auth_perm)


def reapplabel_group_permissions_down(apps, schema_editor):
    reapplabel_group_permissions_up(apps, schema_editor, up=False)


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0080_proxy_group_model'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0012_alter_user_first_name_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('guardian', '0002_generic_permissions_index'),
    ]

    operations = [
        migrations.RunPython(reapplabel_group_permissions_up, reverse_code=reapplabel_group_permissions_down, elidable=True),
    ]



# Generated by Django 2.2.2 on 2019-06-28 13:08

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.models.content
import pulpcore.app.models.fields
import pulpcore.app.models.publication
import pulpcore.app.models.task
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Artifact',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('file', pulpcore.app.models.fields.ArtifactFileField(max_length=255, upload_to=pulpcore.app.models.content.Artifact.storage_path)),
                ('size', models.IntegerField()),
                ('md5', models.CharField(db_index=True, max_length=32)),
                ('sha1', models.CharField(db_index=True, max_length=40)),
                ('sha224', models.CharField(db_index=True, max_length=56)),
                ('sha256', models.CharField(db_index=True, max_length=64, unique=True)),
                ('sha384', models.CharField(db_index=True, max_length=96, unique=True)),
                ('sha512', models.CharField(db_index=True, max_length=128, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Content',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('_type', models.TextField(default=None)),
            ],
            options={
                'verbose_name_plural': 'content',
            },
            bases=(models.Model, pulpcore.app.models.content.QueryMixin),
        ),
        migrations.CreateModel(
            name='ContentAppStatus',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('last_heartbeat', models.DateTimeField(auto_now=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ContentArtifact',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('relative_path', models.CharField(max_length=255)),
                ('artifact', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='core.Artifact')),
                ('content', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Content')),
            ],
            options={
                'unique_together': {('content', 'relative_path')},
            },
            bases=(models.Model, pulpcore.app.models.content.QueryMixin),
        ),
        migrations.CreateModel(
            name='ContentGuard',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('_type', models.TextField(default=None)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('description', models.TextField(null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Exporter',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('_type', models.TextField(default=None)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('last_export', models.DateTimeField(null=True)),
            ],
            options={
                'default_related_name': 'exporters',
            },
        ),
        migrations.CreateModel(
            name='Publisher',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('_type', models.TextField(default=None)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
            ],
            options={
                'default_related_name': 'publishers',
            },
        ),
        migrations.CreateModel(
            name='Remote',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('_type', models.TextField(default=None)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('url', models.TextField()),
                ('ssl_ca_certificate', models.TextField(null=True)),
                ('ssl_client_certificate', models.TextField(null=True)),
                ('ssl_client_key', models.TextField(null=True)),
                ('ssl_validation', models.BooleanField(default=True)),
                ('proxy_url', models.TextField(null=True)),
                ('username', models.TextField(null=True)),
                ('password', models.TextField(null=True)),
                ('download_concurrency', models.PositiveIntegerField(default=20)),
                ('policy', models.TextField(choices=[('immediate', 'When syncing, download all metadata and content now.'), ('on_demand', 'When syncing, download metadata, but do not download content now. Instead, download content as clients request it, and save it in Pulp to be served for future client requests.'), ('streamed', 'When syncing, download metadata, but do not download content now. Instead,download content as clients request it, but never save it in Pulp. This causes future requests for that same content to have to be downloaded again.')], default='immediate')),
            ],
            options={
                'default_related_name': 'remotes',
            },
        ),
        migrations.CreateModel(
            name='Repository',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('description', models.TextField(null=True)),
                ('last_version', models.PositiveIntegerField(default=0)),
            ],
            options={
                'verbose_name_plural': 'repositories',
            },
        ),
        migrations.CreateModel(
            name='RepositoryVersion',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('number', models.PositiveIntegerField(db_index=True)),
                ('complete', models.BooleanField(db_index=True, default=False)),
                ('base_version', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='versions', to='core.RepositoryVersion')),
                ('repository', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='versions', to='core.Repository')),
            ],
            options={
                'ordering': ('number',),
                'get_latest_by': 'number',
                'default_related_name': 'versions',
                'unique_together': {('repository', 'number')},
            },
        ),
        migrations.CreateModel(
            name='ReservedResource',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('resource', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('state', models.TextField(choices=[('waiting', 'Waiting'), ('skipped', 'Skipped'), ('running', 'Running'), ('completed', 'Completed'), ('failed', 'Failed'), ('canceled', 'Canceled')])),
                ('name', models.CharField(max_length=255)),
                ('started_at', models.DateTimeField(null=True)),
                ('finished_at', models.DateTimeField(null=True)),
                ('non_fatal_errors', django.contrib.postgres.fields.JSONField(default=list)),
                ('error', django.contrib.postgres.fields.JSONField(null=True)),
                ('parent', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='spawned_tasks', to='core.Task')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Upload',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('file', models.FileField(max_length=255, upload_to='')),
                ('size', models.BigIntegerField()),
                ('completed', models.DateTimeField(null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Worker',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('last_heartbeat', models.DateTimeField(auto_now=True)),
                ('gracefully_stopped', models.BooleanField(default=False)),
                ('cleaned_up', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UploadChunk',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('offset', models.BigIntegerField()),
                ('size', models.BigIntegerField()),
                ('upload', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='chunks', to='core.Upload')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TaskReservedResource',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('resource', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.ReservedResource')),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='core.Task')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='task',
            name='worker',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tasks', to='core.Worker'),
        ),
        migrations.AddField(
            model_name='reservedresource',
            name='tasks',
            field=models.ManyToManyField(related_name='reserved_resources', through='core.TaskReservedResource', to='core.Task'),
        ),
        migrations.AddField(
            model_name='reservedresource',
            name='worker',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reservations', to='core.Worker'),
        ),
        migrations.CreateModel(
            name='RepositoryVersionContentDetails',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('count_type', models.CharField(choices=[('A', 'added'), ('P', 'present'), ('R', 'removed')], max_length=1)),
                ('content_type', models.TextField()),
                ('count', models.IntegerField()),
                ('repository_version', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='counts', to='core.RepositoryVersion')),
            ],
        ),
        migrations.CreateModel(
            name='RepositoryContent',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('content', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='version_memberships', to='core.Content')),
                ('repository', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Repository')),
                ('version_added', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='added_memberships', to='core.RepositoryVersion')),
                ('version_removed', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='removed_memberships', to='core.RepositoryVersion')),
            ],
            options={
                'unique_together': {('repository', 'content', 'version_added'), ('repository', 'content', 'version_removed')},
            },
        ),
        migrations.AddField(
            model_name='repository',
            name='content',
            field=models.ManyToManyField(related_name='repositories', through='core.RepositoryContent', to='core.Content'),
        ),
        migrations.CreateModel(
            name='Publication',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('_type', models.TextField(default=None)),
                ('complete', models.BooleanField(db_index=True, default=False)),
                ('pass_through', models.BooleanField(default=False)),
                ('publisher', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Publisher')),
                ('repository_version', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.RepositoryVersion')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProgressReport',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('message', models.TextField()),
                ('state', models.TextField(choices=[('waiting', 'Waiting'), ('skipped', 'Skipped'), ('running', 'Running'), ('completed', 'Completed'), ('failed', 'Failed'), ('canceled', 'Canceled')], default='waiting')),
                ('total', models.IntegerField(null=True)),
                ('done', models.IntegerField(default=0)),
                ('suffix', models.TextField(null=True)),
                ('task', models.ForeignKey(default=pulpcore.app.models.task.Task.current, on_delete=django.db.models.deletion.CASCADE, related_name='progress_reports', to='core.Task')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CreatedResource',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('object_id', models.UUIDField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('task', models.ForeignKey(default=pulpcore.app.models.task.Task.current, on_delete=django.db.models.deletion.CASCADE, related_name='created_resources', to='core.Task')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='content',
            name='_artifacts',
            field=models.ManyToManyField(through='core.ContentArtifact', to='core.Artifact'),
        ),
        migrations.CreateModel(
            name='BaseDistribution',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('_type', models.TextField(default=None)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('base_path', models.CharField(max_length=255, unique=True)),
                ('content_guard', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.ContentGuard')),
                ('remote', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.Remote')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProgressBar',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('core.progressreport',),
        ),
        migrations.CreateModel(
            name='ProgressSpinner',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('core.progressreport',),
        ),
        migrations.CreateModel(
            name='RemoteArtifact',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('url', models.TextField(validators=[django.core.validators.URLValidator])),
                ('size', models.IntegerField(null=True)),
                ('md5', models.CharField(max_length=32, null=True)),
                ('sha1', models.CharField(max_length=40, null=True)),
                ('sha224', models.CharField(max_length=56, null=True)),
                ('sha256', models.CharField(max_length=64, null=True)),
                ('sha384', models.CharField(max_length=96, null=True)),
                ('sha512', models.CharField(max_length=128, null=True)),
                ('content_artifact', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.ContentArtifact')),
                ('remote', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Remote')),
            ],
            options={
                'unique_together': {('content_artifact', 'remote')},
            },
            bases=(models.Model, pulpcore.app.models.content.QueryMixin),
        ),
        migrations.CreateModel(
            name='PublishedMetadata',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('relative_path', models.CharField(max_length=255)),
                ('file', models.FileField(max_length=255, upload_to='')),
                ('publication', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='published_metadata', to='core.Publication')),
            ],
            options={
                'default_related_name': 'published_metadata',
                'unique_together': {('publication', 'relative_path'), ('publication', 'file')},
            },
        ),
        migrations.CreateModel(
            name='PublishedArtifact',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('relative_path', models.CharField(max_length=255)),
                ('content_artifact', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='published_artifact', to='core.ContentArtifact')),
                ('publication', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='published_artifact', to='core.Publication')),
            ],
            options={
                'default_related_name': 'published_artifact',
                'unique_together': {('publication', 'content_artifact'), ('publication', 'relative_path')},
            },
        ),
    ]



from unittest.mock import patch

from django.test import TestCase
from rest_framework.serializers import ValidationError

from pulpcore.app.tasks.importer import _check_versions


class TestObject:
    version = "1.2.3"  # Every component is vers 1.2.3


class TestCheckVersions(TestCase):
    @patch("pulpcore.app.tasks.importer.get_distribution", return_value=TestObject())
    def test_vers_check(self, mock_get_distribution):
        export_json = [{"component": "xyz", "version": "1.2.3"}]
        _check_versions(export_json)

        export_json = [{"component": "xy", "version": "1.2"}]
        _check_versions(export_json)

        export_json = [{"component": "x_noty_z", "version": "1.4.3"}]
        with self.assertRaises(ValidationError):
            _check_versions(export_json)

        export_json = [{"component": "notx_y_z", "version": "2.2.3"}]
        with self.assertRaises(ValidationError):
            _check_versions(export_json)



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

from pulpcore.app.models import BaseModel, Group


class Role(BaseModel):
    """
    A model for the "role" part in RBAC.

    Fields:

        name (models.CharField): Unique name of the role.
        locked (models.BooleanField): Indicator for plugin managed role.

    Relations:

        permissions (models.ManyToManyField): Permissions to be granted via this role.
    """

    name = models.CharField(max_length=128, db_index=True, unique=True)
    description = models.TextField(null=True)
    locked = models.BooleanField(default=False)
    permissions = models.ManyToManyField(Permission)


class UserRole(BaseModel):
    """
    Join table for user to role associations with optional content object.

    Relations:

        user (models.ForeignKey): User to grant permissions to.
        role (models.ForeignKey): Role to select granted permissions from.
        content_object (GenericForeignKey): Optional object to assert permissions on.
    """

    user = models.ForeignKey(
        get_user_model(), related_name="object_roles", on_delete=models.CASCADE
    )
    role = models.ForeignKey(Role, related_name="object_users", on_delete=models.CASCADE)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True)
    object_id = models.CharField(max_length=255, null=True)
    content_object = GenericForeignKey("content_type", "object_id", for_concrete_model=False)

    class Meta:
        unique_together = (("user", "role", "content_type", "object_id"),)
        indexes = [models.Index(fields=["content_type", "object_id"])]


class GroupRole(BaseModel):
    """
    Join table for group to role associations with optional content object.

    Relations:

        group (models.ForeignKey): Group to grant permissions to.
        role (models.ForeignKey): Role to select granted permissions from.
        content_object (GenericForeignKey): Optional object to assert permissions on.
    """

    group = models.ForeignKey(Group, related_name="object_roles", on_delete=models.CASCADE)
    role = models.ForeignKey(Role, related_name="object_groups", on_delete=models.CASCADE)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True)
    object_id = models.CharField(max_length=255, null=True)
    content_object = GenericForeignKey("content_type", "object_id", for_concrete_model=False)

    class Meta:
        unique_together = (("group", "role", "content_type", "object_id"),)
        indexes = [models.Index(fields=["content_type", "object_id"])]



"""
Repository related Django models.
"""
from contextlib import suppress
from gettext import gettext as _
from os import path
from collections import defaultdict
import logging

import django
from asyncio_throttle import Throttler
from dynaconf import settings
from django.core.validators import MinValueValidator
from django.db import models, transaction
from django.db.models import Q
from django.urls import reverse
from django_lifecycle import AFTER_UPDATE, BEFORE_DELETE, hook
from rest_framework.exceptions import APIException

from pulpcore.app.util import batch_qs, get_view_name_for_model
from pulpcore.constants import ALL_KNOWN_CONTENT_CHECKSUMS
from pulpcore.download.factory import DownloaderFactory
from pulpcore.exceptions import ResourceImmutableError

from pulpcore.cache import Cache

from .base import MasterModel, BaseModel
from .content import Artifact, Content
from .fields import EncryptedTextField
from .task import CreatedResource, Task


_logger = logging.getLogger(__name__)


class Repository(MasterModel):
    """
    Collection of content.

    Fields:

        name (models.TextField): The repository name.
        description (models.TextField): An optional description.
        next_version (models.PositiveIntegerField): A record of the next version number to be
            created.
        retain_repo_versions (models.PositiveIntegerField): Number of repo versions to keep
        user_hidden (models.BooleanField): Whether to expose this repo to users via the API

    Relations:

        content (models.ManyToManyField): Associated content.
        remote (models.ForeignKeyField): Associated remote
    """

    TYPE = "repository"
    CONTENT_TYPES = []
    REMOTE_TYPES = []

    name = models.TextField(db_index=True, unique=True)
    description = models.TextField(null=True)
    next_version = models.PositiveIntegerField(default=0)
    retain_repo_versions = models.PositiveIntegerField(default=None, null=True)
    user_hidden = models.BooleanField(default=False)
    content = models.ManyToManyField(
        "Content", through="RepositoryContent", related_name="repositories"
    )
    remote = models.ForeignKey("Remote", null=True, on_delete=models.SET_NULL)

    class Meta:
        verbose_name_plural = "repositories"

    def on_new_version(self, version):
        """Called after a new repository version has been created.

        Subclasses are expected to override this to do useful things.

        Args:
            version: The new repository version.
        """
        pass

    def save(self, *args, **kwargs):
        """
        Saves Repository model and creates an initial repository version.

        Args:
            args (list): list of positional arguments for Model.save()
            kwargs (dict): dictionary of keyword arguments to pass to Model.save()
        """
        with transaction.atomic():
            adding = self._state.adding
            super().save(*args, **kwargs)
            if adding:
                self.create_initial_version()

    def create_initial_version(self):
        """
        Create an initial repository version (version 0).

        This method can be overriden by plugins if they require custom logic.
        """
        version = RepositoryVersion(repository=self, number=self.next_version, complete=True)
        self.next_version += 1
        self.save()
        version.save()

    def new_version(self, base_version=None):
        """
        Create a new RepositoryVersion for this Repository

        Creation of a RepositoryVersion should be done in a RQ Job.

        Args:
            repository (pulpcore.app.models.Repository): to create a new version of
            base_version (pulpcore.app.models.RepositoryVersion): an optional repository version
                whose content will be used as the set of content for the new version

        Returns:
            pulpcore.app.models.RepositoryVersion: The Created RepositoryVersion
        """
        with transaction.atomic():
            latest_version = self.versions.latest()
            if not latest_version.complete:
                latest_version.delete()

            version = RepositoryVersion(
                repository=self, number=int(self.next_version), base_version=base_version
            )
            version.save()

            if base_version:
                # first remove the content that isn't in the base version
                version.remove_content(version.content.exclude(pk__in=base_version.content))
                # now add any content that's in the base_version but not in version
                version.add_content(base_version.content.exclude(pk__in=version.content))

            if Task.current() and not self.user_hidden:
                resource = CreatedResource(content_object=version)
                resource.save()

            self.invalidate_cache()

            return version

    def initialize_new_version(self, new_version):
        """
        Initialize the new RepositoryVersion with plugin-provided code.

        This method should be overridden by plugin writers for an opportunity for plugin input. This
        method is intended to be called with the incomplete
        :class:`pulpcore.app.models.RepositoryVersion` to validate or modify the content.

        This method does not adjust the value of complete, or save the `RepositoryVersion` itself.
        Its intent is to allow the plugin writer an opportunity for plugin input before any other
        actions performed on the new `RepositoryVersion`.

        Args:
            new_version (pulpcore.app.models.RepositoryVersion): The incomplete RepositoryVersion to
                finalize.

        """
        pass

    def finalize_new_version(self, new_version):
        """
        Finalize the incomplete RepositoryVersion with plugin-provided code.

        This method should be overridden by plugin writers for an opportunity for plugin input. This
        method is intended to be called with the incomplete
        :class:`pulpcore.app.models.RepositoryVersion` to validate or modify the content.

        This method does not adjust the value of complete, or save the `RepositoryVersion` itself.
        Its intent is to allow the plugin writer an opportunity for plugin input before pulpcore
        marks the `RepositoryVersion` as complete.

        Args:
            new_version (pulpcore.app.models.RepositoryVersion): The incomplete RepositoryVersion to
                finalize.

        Returns:

        """
        pass

    def latest_version(self):
        """
        Get the latest RepositoryVersion on a repository

        Args:
            repository (pulpcore.app.models.Repository): to get the latest version of

        Returns:
            pulpcore.app.models.RepositoryVersion: The latest RepositoryVersion

        """
        with suppress(RepositoryVersion.DoesNotExist):
            model = self.versions.complete().latest()
            return model

    def natural_key(self):
        """
        Get the model's natural key.

        :return: The model's natural key.
        :rtype: tuple
        """
        return (self.name,)

    @staticmethod
    def artifacts_for_version(version):
        """
        Return the artifacts for a repository version.

        Provides a method that plugins can override since RepositoryVersions aren't typed.

        Args:
            version (pulpcore.app.models.RepositoryVersion): to get the artifacts for

        Returns:
            django.db.models.QuerySet: The artifacts that are contained within this version.
        """
        return Artifact.objects.filter(content__pk__in=version.content)

    @hook(AFTER_UPDATE, when="retain_repo_versions", has_changed=True)
    def _cleanup_old_versions_hook(self):
        # Do not attempt to clean up anything, while there is a transaction involving repo versions
        # still in flight.
        transaction.on_commit(self.cleanup_old_versions)

    def cleanup_old_versions(self):
        """Cleanup old repository versions based on retain_repo_versions."""
        # I am still curious how, but it was reported that this state can happen in day to day
        # operations but its easy to reproduce manually in the pulpcore shell:
        # https://github.com/pulp/pulpcore/issues/2268
        if self.versions.filter(complete=False).exists():
            raise RuntimeError(
                _("Attempt to cleanup old versions, while a new version is in flight.")
            )
        if self.retain_repo_versions:
            # Consider only completed versions for cleanup
            for version in self.versions.complete().order_by("-number")[
                self.retain_repo_versions :
            ]:
                _logger.info(
                    _("Deleting repository version {} due to version retention limit.").format(
                        version
                    )
                )
                version.delete()

    @hook(BEFORE_DELETE)
    def invalidate_cache(self, everything=False):
        """Invalidates the cache if repository is present."""
        if settings.CACHE_ENABLED:
            distributions = self.distributions.all()
            if everything:
                from .publication import Distribution, Publication

                versions = self.versions.all()
                pubs = Publication.objects.filter(repository_version__in=versions, complete=True)
                distributions |= Distribution.objects.filter(publication__in=pubs)
                distributions |= Distribution.objects.filter(repository_version__in=versions)
            if distributions.exists():
                base_paths = distributions.values_list("base_path", flat=True)
                if base_paths:
                    Cache().delete(base_key=base_paths)
                # Could do preloading here for immediate artifacts with artifacts_for_version


class Remote(MasterModel):
    """
    A remote source for content.

    This is meant to be subclassed by plugin authors as an opportunity to provide plugin-specific
    persistent data attributes for a plugin remote subclass.

    This object is a Django model that inherits from :class: `pulpcore.app.models.Remote` which
    provides the platform persistent attributes for a remote object. Plugin authors can add
    additional persistent remote data by subclassing this object and adding Django fields. We
    defer to the Django docs on extending this model definition with additional fields.

    Validation of the remote is done at the API level by a plugin defined subclass of
    :class: `pulpcore.plugin.serializers.repository.RemoteSerializer`.

    Fields:

        name (models.TextField): The remote name.
        url (models.TextField): The URL of an external content source.
        ca_cert (models.TextField): A PEM encoded CA certificate used to validate the
            server certificate presented by the external source.
        client_cert (models.TextField): A PEM encoded client certificate used
            for authentication.
        client_key (models.TextField): A PEM encoded private key used for authentication.
        tls_validation (models.BooleanField): If True, TLS peer validation must be performed.
        proxy_url (models.TextField): The optional proxy URL.
            Format: scheme://host:port
        proxy_username (models.TextField): The optional username to authenticate with the proxy.
        proxy_password (models.TextField): The optional password to authenticate with the proxy.
        username (models.TextField): The username to be used for authentication when syncing.
        password (models.TextField): The password to be used for authentication when syncing.
        download_concurrency (models.PositiveIntegerField): Total number of
            simultaneous connections allowed to any remote during a sync.
        policy (models.TextField): The policy to use when downloading content.
        total_timeout (models.FloatField): Value for aiohttp.ClientTimeout.total on connections
        connect_timeout (models.FloatField): Value for aiohttp.ClientTimeout.connect
        sock_connect_timeout (models.FloatField): Value for aiohttp.ClientTimeout.sock_connect
        sock_read_timeout (models.FloatField): Value for aiohttp.ClientTimeout.sock_read
        headers (models.JSONField): Headers set on the aiohttp.ClientSession
        rate_limit (models.IntegerField): Limits requests per second for each concurrent downloader
    """

    TYPE = "remote"

    # Constants for the ChoiceField 'policy'
    IMMEDIATE = "immediate"
    ON_DEMAND = "on_demand"
    STREAMED = "streamed"

    DEFAULT_DOWNLOAD_CONCURRENCY = 10
    DEFAULT_MAX_RETRIES = 3

    POLICY_CHOICES = (
        (IMMEDIATE, "When syncing, download all metadata and content now."),
        (
            ON_DEMAND,
            "When syncing, download metadata, but do not download content now. Instead, "
            "download content as clients request it, and save it in Pulp to be served for "
            "future client requests.",
        ),
        (
            STREAMED,
            "When syncing, download metadata, but do not download content now. Instead,"
            "download content as clients request it, but never save it in Pulp. This causes "
            "future requests for that same content to have to be downloaded again.",
        ),
    )

    name = models.TextField(db_index=True, unique=True)

    url = models.TextField()

    ca_cert = models.TextField(null=True)
    client_cert = models.TextField(null=True)
    client_key = EncryptedTextField(null=True)
    tls_validation = models.BooleanField(default=True)

    username = EncryptedTextField(null=True)
    password = EncryptedTextField(null=True)

    proxy_url = models.TextField(null=True)
    proxy_username = EncryptedTextField(null=True)
    proxy_password = EncryptedTextField(null=True)

    download_concurrency = models.PositiveIntegerField(
        null=True, validators=[MinValueValidator(1, "Download concurrency must be at least 1")]
    )
    max_retries = models.PositiveIntegerField(null=True)
    policy = models.TextField(choices=POLICY_CHOICES, default=IMMEDIATE)

    total_timeout = models.FloatField(
        null=True, validators=[MinValueValidator(0.0, "Timeout must be >= 0")]
    )
    connect_timeout = models.FloatField(
        null=True, validators=[MinValueValidator(0.0, "Timeout must be >= 0")]
    )
    sock_connect_timeout = models.FloatField(
        null=True, validators=[MinValueValidator(0.0, "Timeout must be >= 0")]
    )
    sock_read_timeout = models.FloatField(
        null=True, validators=[MinValueValidator(0.0, "Timeout must be >= 0")]
    )
    headers = models.JSONField(blank=True, null=True)
    rate_limit = models.IntegerField(null=True)

    @property
    def download_factory(self):
        """
        Return the DownloaderFactory which can be used to generate asyncio capable downloaders.

        Upon first access, the DownloaderFactory is instantiated and saved internally.

        Plugin writers are expected to override when additional configuration of the
        DownloaderFactory is needed.

        Returns:
            DownloadFactory: The instantiated DownloaderFactory to be used by
                get_downloader().
        """
        try:
            return self._download_factory
        except AttributeError:
            self._download_factory = DownloaderFactory(self)
            return self._download_factory

    @property
    def download_throttler(self):
        """
        Return the Throttler which can be used to rate limit downloaders.

        Upon first access, the Throttler is instantiated and saved internally.
        Plugin writers are expected to override when additional configuration of the
        DownloaderFactory is needed.

        Returns:
            Throttler: The instantiated Throttler to be used by get_downloader()

        """
        try:
            return self._download_throttler
        except AttributeError:
            if self.rate_limit:
                self._download_throttler = Throttler(rate_limit=self.rate_limit)
                return self._download_throttler

    def get_downloader(self, remote_artifact=None, url=None, download_factory=None, **kwargs):
        """
        Get a downloader from either a RemoteArtifact or URL that is configured with this Remote.

        This method accepts either `remote_artifact` or `url` but not both. At least one is
        required. If neither or both are passed a ValueError is raised.

        Plugin writers are expected to override when additional configuration is needed or when
        another class of download is required.

        Args:
            remote_artifact (:class:`~pulpcore.app.models.RemoteArtifact`): The RemoteArtifact to
                download.
            url (str): The URL to download.
            download_factory (:class:`~pulpcore.plugin.download.DownloadFactory`): The download
                factory to be used.
            kwargs (dict): This accepts the parameters of
                :class:`~pulpcore.plugin.download.BaseDownloader`.

        Raises:
            ValueError: If neither remote_artifact and url are passed, or if both are passed.

        Returns:
            subclass of :class:`~pulpcore.plugin.download.BaseDownloader`: A downloader that
            is configured with the remote settings.
        """
        if remote_artifact and url:
            raise ValueError(_("get_downloader() cannot accept both 'remote_artifact' and 'url'."))
        if remote_artifact is None and url is None:
            raise ValueError(_("get_downloader() requires either 'remote_artifact' and 'url'."))
        if remote_artifact:
            url = remote_artifact.url
            expected_digests = {}
            for digest_name in ALL_KNOWN_CONTENT_CHECKSUMS:
                digest_value = getattr(remote_artifact, digest_name)
                if digest_value:
                    expected_digests[digest_name] = digest_value
            if expected_digests:
                kwargs["expected_digests"] = expected_digests
            if remote_artifact.size:
                kwargs["expected_size"] = remote_artifact.size
        if download_factory is None:
            download_factory = self.download_factory
        return download_factory.build(url, **kwargs)

    def get_remote_artifact_url(self, relative_path=None, request=None):
        """
        Get the full URL for a RemoteArtifact from relative path and request.

        This method returns the URL for a RemoteArtifact by concatenating the Remote's url and the
        relative path. Plugin writers are expected to override this method when a more complex
        algorithm is needed to determine the full URL.

        Args:
            relative_path (str): The relative path of a RemoteArtifact
            request (aiohttp.web.Request): The request object for this relative path

        Raises:
            ValueError: If relative_path starts with a '/'.

        Returns:
            str: A URL for a RemoteArtifact available at the Remote.
        """
        if path.isabs(relative_path):
            raise ValueError(_("Relative path can't start with '/'. {0}").format(relative_path))
        return path.join(self.url, relative_path)

    def get_remote_artifact_content_type(self, relative_path=None):
        """
        Get the type of content that should be available at the relative path.

        Plugin writers are expected to implement this method.

        Args:
            relative_path (str): The relative path of a RemoteArtifact

        Returns:
            Class: The Class of the content type that should be available at the relative path.
        """
        raise NotImplementedError()

    @hook(BEFORE_DELETE)
    def invalidate_cache(self):
        """Invalidates the cache if remote is present."""
        if settings.CACHE_ENABLED:
            base_paths = self.distribution_set.values_list("base_path", flat=True)
            if base_paths:
                Cache().delete(base_key=base_paths)

    class Meta:
        default_related_name = "remotes"


class RepositoryContent(BaseModel):
    """
    Association between a repository and its contained content.

    Fields:

        created (models.DatetimeField): When the association was created.

    Relations:

        content (models.ForeignKey): The associated content.
        repository (models.ForeignKey): The associated repository.
        version_added (models.ForeignKey): The RepositoryVersion which added the referenced
            Content.
        version_removed (models.ForeignKey): The RepositoryVersion which removed the referenced
            Content.
    """

    content = models.ForeignKey(
        "Content", on_delete=models.CASCADE, related_name="version_memberships"
    )
    repository = models.ForeignKey(Repository, on_delete=models.CASCADE)
    version_added = models.ForeignKey(
        "RepositoryVersion", related_name="added_memberships", on_delete=models.CASCADE
    )
    version_removed = models.ForeignKey(
        "RepositoryVersion", null=True, related_name="removed_memberships", on_delete=models.CASCADE
    )

    class Meta:
        unique_together = (
            ("repository", "content", "version_added"),
            ("repository", "content", "version_removed"),
        )


class RepositoryVersionQuerySet(models.QuerySet):
    """A queryset that provides repository version filtering methods."""

    def complete(self):
        return self.filter(complete=True)

    def with_content(self, content):
        """
        Filters repository versions that contain the provided content units.

        Args:
            content (django.db.models.QuerySet): query of content

        Returns:
            django.db.models.QuerySet: Repository versions which contains content.
        """
        query = models.Q(pk__in=[])
        repo_content = RepositoryContent.objects.filter(content__pk__in=content)

        for rc in repo_content.iterator():
            filter = models.Q(
                repository__pk=rc.repository.pk,
                number__gte=rc.version_added.number,
            )
            if rc.version_removed:
                filter &= models.Q(number__lt=rc.version_removed.number)

            query |= filter

        return self.filter(query)


class RepositoryVersion(BaseModel):
    """
    A version of a repository's content set.

    Plugin Writers are strongly encouraged to use RepositoryVersion as a context manager to provide
    transactional safety, working directory set up, plugin finalization, and cleaning up the
    database on failures.

    Examples::

        with repository.new_version(repository) as new_version:
            new_version.add_content(content_q)
            new_version.remove_content(content_q)

    Fields:

        number (models.PositiveIntegerField): A positive integer that uniquely identifies a version
            of a specific repository. Each new version for a repo should have this field set to
            1 + the most recent version.
        complete (models.BooleanField): If true, the RepositoryVersion is visible. This field is set
            to true when the task that creates the RepositoryVersion is complete.

    Relations:

        repository (models.ForeignKey): The associated repository.
        base_version (models.ForeignKey): The repository version this was created from.
    """

    objects = RepositoryVersionQuerySet.as_manager()

    repository = models.ForeignKey(Repository, on_delete=models.CASCADE)
    number = models.PositiveIntegerField(db_index=True)
    complete = models.BooleanField(db_index=True, default=False)
    base_version = models.ForeignKey("RepositoryVersion", null=True, on_delete=models.SET_NULL)

    class Meta:
        default_related_name = "versions"
        unique_together = ("repository", "number")
        get_latest_by = "number"
        ordering = ("number",)

    def _content_relationships(self):
        """
        Returns a set of repository_content for a repository version

        Returns:
            django.db.models.QuerySet: The repository_content that is contained within this version.
        """
        return RepositoryContent.objects.filter(
            repository=self.repository, version_added__number__lte=self.number
        ).exclude(version_removed__number__lte=self.number)

    def get_content(self, content_qs=None):
        """
        Returns a set of content for a repository version

        Args:
            content_qs (:class:`django.db.models.QuerySet`): The queryset for Content that will be
                restricted further to the content present in this repository version. If not given,
                ``Content.objects.all()`` is used (to return over all content types present in the
                repository version).

        Returns:
            django.db.models.QuerySet: The content that is contained within this version.

        Examples:
            >>> repository_version = ...
            >>>
            >>> # Return a queryset of File objects in the repository
            >>> repository_version.get_content(content_qs=File.objects)):
        """

        if content_qs is None:
            content_qs = Content.objects

        return content_qs.filter(version_memberships__in=self._content_relationships())

    @property
    def content(self):
        """
        Returns a set of content for a repository version

        Returns:
            django.db.models.QuerySet: The content that is contained within this version.

        Examples:
            >>> repository_version = ...
            >>>
            >>> for content in repository_version.content:
            >>>     content = content.cast()  # optional downcast.
            >>>     ...
            >>>
            >>> for content in FileContent.objects.filter(pk__in=repository_version.content):
            >>>     ...
            >>>
        """

        return self.get_content()

    def content_batch_qs(self, content_qs=None, order_by_params=("pk",), batch_size=1000):
        """
        Generate content batches to efficiently iterate over all content.

        Generates query sets that span the `content_qs` content of the repository
        version. Each yielded query set evaluates to at most `batch_size` content records.
        This is useful to limit the memory footprint when iterating over all content of
        a repository version.

        .. note::

            * This generator is not safe against changes (i.e. add/remove content) during
              the iteration!

            * As the method uses slices internally, the queryset must be ordered to yield
              stable results. By default, it is ordered by primary key.

        Args:
            content_qs (:class:`django.db.models.QuerySet`): The queryset for Content that will be
                restricted further to the content present in this repository version. If not given,
                ``Content.objects.all()`` is used (to iterate over all content present in the
                repository version). A plugin may want to use a specific subclass of
                :class:`~pulpcore.plugin.models.Content` or use e.g. ``filter()`` to select
                a subset of the repository version's content.
            order_by_params (tuple of str): The parameters for the ``order_by`` clause
                for the content. The Default is ``("pk",)``. This needs to
                specify a stable order. For example, if you want to iterate by
                decreasing creation time stamps use ``("-pulp_created", "pk")`` to
                ensure that content records are still sorted by primary key even
                if their creation timestamp happens to be equal.
            batch_size (int): The maximum batch size.

        Yields:
            :class:`django.db.models.QuerySet`: A QuerySet representing a slice of the content.

        Example:
            The following code could be used to loop over all ``FileContent`` in
            ``repository_version``. It prefetches the related
            :class:`~pulpcore.plugin.models.ContentArtifact` instances for every batch::

                repository_version = ...

                batch_generator = repository_version.content_batch_qs(
                    content_class=FileContent.objects.all()
                )
                for content_batch_qs in batch_generator:
                    content_batch_qs.prefetch_related("contentartifact_set")
                    for content in content_batch_qs:
                        ...

        """
        version_content_qs = self.get_content(content_qs).order_by(*order_by_params)
        yield from batch_qs(version_content_qs, batch_size=batch_size)

    @property
    def artifacts(self):
        """
        Returns a set of artifacts for a repository version.

        Returns:
            django.db.models.QuerySet: The artifacts that are contained within this version.
        """
        return self.repository.cast().artifacts_for_version(self)

    def added(self, base_version=None):
        """
        Args:
            base_version (pulpcore.app.models.RepositoryVersion): an optional repository version

        Returns:
            QuerySet: The Content objects that were added by this version.
        """
        if not base_version:
            return Content.objects.filter(version_memberships__version_added=self)

        return Content.objects.filter(
            version_memberships__in=self._content_relationships()
        ).exclude(version_memberships__in=base_version._content_relationships())

    def removed(self, base_version=None):
        """
        Args:
            base_version (pulpcore.app.models.RepositoryVersion): an optional repository version

        Returns:
            QuerySet: The Content objects that were removed by this version.
        """
        if not base_version:
            return Content.objects.filter(version_memberships__version_removed=self)

        return Content.objects.filter(
            version_memberships__in=base_version._content_relationships()
        ).exclude(version_memberships__in=self._content_relationships())

    def contains(self, content):
        """
        Check whether a content exists in this repository version's set of content

        Returns:
            bool: True if the repository version contains the content, False otherwise
        """
        return self.content.filter(pk=content.pk).exists()

    def add_content(self, content):
        """
        Add a content unit to this version.

        Args:
           content (django.db.models.QuerySet): Set of Content to add

        Raise:
            pulpcore.exception.ResourceImmutableError: if add_content is called on a
                complete RepositoryVersion
        """

        if self.complete:
            raise ResourceImmutableError(self)

        repo_content = []
        to_add = set(content.exclude(pk__in=self.content).values_list("pk", flat=True))

        # Normalize representation if content has already been removed in this version and
        # is re-added: Undo removal by setting version_removed to None.
        for removed in batch_qs(self.removed().order_by("pk").values_list("pk", flat=True)):
            to_readd = to_add.intersection(set(removed))
            if to_readd:
                RepositoryContent.objects.filter(
                    content__in=to_readd, repository=self.repository, version_removed=self
                ).update(version_removed=None)
                to_add = to_add - to_readd

        for content_pk in to_add:
            repo_content.append(
                RepositoryContent(
                    repository=self.repository, content_id=content_pk, version_added=self
                )
            )

        RepositoryContent.objects.bulk_create(repo_content)

    def remove_content(self, content):
        """
        Remove content from the repository.

        Args:
            content (django.db.models.QuerySet): Set of Content to remove

        Raise:
            pulpcore.exception.ResourceImmutableError: if remove_content is called on a
                complete RepositoryVersion
        """

        if self.complete:
            raise ResourceImmutableError(self)

        if not content or not content.count():
            return

        # Normalize representation if content has already been added in this version.
        # Undo addition by deleting the RepositoryContent.
        RepositoryContent.objects.filter(
            repository=self.repository,
            content_id__in=content,
            version_added=self,
            version_removed=None,
        ).delete()

        q_set = RepositoryContent.objects.filter(
            repository=self.repository, content_id__in=content, version_removed=None
        )
        q_set.update(version_removed=self)

    def set_content(self, content):
        """
        Sets the repo version content by calling remove_content() then add_content().

        Args:
            content (django.db.models.QuerySet): Set of desired content

        Raise:
            pulpcore.exception.ResourceImmutableError: if set_content is called on a
                complete RepositoryVersion
        """
        self.remove_content(self.content.exclude(pk__in=content))
        self.add_content(content.exclude(pk__in=self.content))

    def next(self):
        """
        Returns:
            pulpcore.app.models.RepositoryVersion: The next complete RepositoryVersion for the same
                repository.
        Raises:
            RepositoryVersion.DoesNotExist: if there is not a RepositoryVersion for the same
                repository and with a higher "number".
        """
        try:
            return (
                self.repository.versions.complete()
                .filter(number__gt=self.number)
                .order_by("number")[0]
            )
        except IndexError:
            raise self.DoesNotExist

    def previous(self):
        """
        Returns:
            pulpcore.app.models.RepositoryVersion: The previous complete RepositoryVersion for the
                same repository.

        Raises:
            RepositoryVersion.DoesNotExist: if there is not a RepositoryVersion for the same
                repository and with a lower "number".
        """
        try:
            return (
                self.repository.versions.complete()
                .filter(number__lt=self.number)
                .order_by("-number")[0]
            )
        except IndexError:
            raise self.DoesNotExist

    def _squash(self, repo_relations, next_version):
        """
        Squash a complete repo version into the next version
        """
        # delete any relationships added in the version being deleted and removed in the next one.
        repo_relations.filter(version_added=self, version_removed=next_version).delete()

        # If the same content is deleted in version, but added back in next_version then:
        # - set version_removed field in relation to version_removed of the relation adding
        #   the content in next version because the content can be removed again after the
        #   next_version
        # - and remove relation adding the content in next_version
        content_added = repo_relations.filter(version_added=next_version).values_list("content_id")

        # use list() to force the evaluation of the queryset, otherwise queryset is affected
        # by the update() operation before delete() is ran
        content_removed_and_readded = list(
            repo_relations.filter(version_removed=self, content_id__in=content_added).values_list(
                "content_id"
            )
        )

        repo_contents_readded_in_next_version = repo_relations.filter(
            version_added=next_version, content_id__in=content_removed_and_readded
        )

        # Since the readded contents can be removed again by any subsequent version after the
        # next version. Get the mapping of readded contents and their versions removed to use
        # later. The version removed id will be None if a content is not removed.
        version_removed_id_content_id_map = defaultdict(list)
        for readded_repo_content in repo_contents_readded_in_next_version.iterator():
            version_removed_id_content_id_map[readded_repo_content.version_removed_id].append(
                readded_repo_content.content_id
            )

        repo_contents_readded_in_next_version.delete()

        # Update the version removed of the readded contents
        for version_removed_id, content_ids in version_removed_id_content_id_map.items():
            repo_relations.filter(version_removed=self, content_id__in=content_ids).update(
                version_removed_id=version_removed_id
            )

        # "squash" by moving other additions and removals forward to the next version
        repo_relations.filter(version_added=self).update(version_added=next_version)
        repo_relations.filter(version_removed=self).update(version_removed=next_version)

        # Update next version's counts as they have been modified
        next_version._compute_counts()

    def delete(self, **kwargs):
        """
        Deletes a RepositoryVersion

        If RepositoryVersion is complete and has a successor, squash RepositoryContent changes into
        the successor. If version is incomplete, delete and and clean up RepositoryContent,
        CreatedResource, and Repository objects.

        Deletion of a complete RepositoryVersion should be done in a RQ Job.
        """
        if self.complete:
            if self.repository.versions.complete().count() <= 1:
                raise APIException(_("Attempt to delete the last remaining version."))
            if settings.CACHE_ENABLED:
                base_paths = self.distribution_set.values_list("base_path", flat=True)
                if base_paths:
                    Cache().delete(base_key=base_paths)

            # Handle the manipulation of the repository version content and its final deletion in
            # the same transaction.
            with transaction.atomic():
                repo_relations = RepositoryContent.objects.filter(
                    repository=self.repository
                ).select_for_update()
                try:
                    next_version = self.next()
                    self._squash(repo_relations, next_version)

                except RepositoryVersion.DoesNotExist:
                    # version is the latest version so simply update repo contents
                    # and delete the version
                    repo_relations.filter(version_added=self).delete()
                    repo_relations.filter(version_removed=self).update(version_removed=None)

                if repo_relations.filter(Q(version_added=self) | Q(version_removed=self)).exists():
                    raise RuntimeError(
                        _("Some repo relations of this version were not translated.")
                    )
                super().delete(**kwargs)

        else:
            with transaction.atomic():
                RepositoryContent.objects.filter(version_added=self).delete()
                RepositoryContent.objects.filter(version_removed=self).update(version_removed=None)
                CreatedResource.objects.filter(object_id=self.pk).delete()
                super().delete(**kwargs)

    def _compute_counts(self):
        """
        Compute and save content unit counts by type.

        Count records are stored as :class:`~pulpcore.app.models.RepositoryVersionContentDetails`.
        This method deletes existing :class:`~pulpcore.app.models.RepositoryVersionContentDetails`
        objects and makes new ones with each call.
        """
        with transaction.atomic():
            counts_list = []
            for value, name in RepositoryVersionContentDetails.COUNT_TYPE_CHOICES:
                RepositoryVersionContentDetails.objects.filter(repository_version=self).delete()
                if value == RepositoryVersionContentDetails.ADDED:
                    qs = self.added()
                elif value == RepositoryVersionContentDetails.PRESENT:
                    qs = self.content
                elif value == RepositoryVersionContentDetails.REMOVED:
                    qs = self.removed()
                annotated = qs.values("pulp_type").annotate(count=models.Count("pulp_type"))
                for item in annotated:
                    count_obj = RepositoryVersionContentDetails(
                        content_type=item["pulp_type"],
                        repository_version=self,
                        count=item["count"],
                        count_type=value,
                    )
                    counts_list.append(count_obj)
            RepositoryVersionContentDetails.objects.bulk_create(counts_list)

    def __enter__(self):
        """
        Create the repository version

        Returns:
            RepositoryVersion: self
        """
        if self.complete:
            raise RuntimeError(
                _("This Repository version is complete. It cannot be modified further.")
            )
        repository = self.repository.cast()
        repository.initialize_new_version(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Finalize and save the RepositoryVersion if no errors are raised, delete it if not
        """
        if exc_value:
            self.delete()
        else:
            try:
                repository = self.repository.cast()
                repository.finalize_new_version(self)
                no_change = not self.added() and not self.removed()
                if no_change:
                    self.delete()
                else:
                    content_types_seen = set(
                        self.content.values_list("pulp_type", flat=True).distinct()
                    )
                    content_types_supported = set(
                        ctype.get_pulp_type() for ctype in repository.CONTENT_TYPES
                    )

                    unsupported_types = content_types_seen - content_types_supported
                    if unsupported_types:
                        raise ValueError(
                            _("Saw unsupported content types {}").format(unsupported_types)
                        )

                    self.complete = True
                    self.repository.next_version = self.number + 1
                    with transaction.atomic():
                        self.repository.save()
                        self.save()
                        self._compute_counts()
                    self.repository.cleanup_old_versions()
                    repository.on_new_version(self)
            except Exception:
                self.delete()
                raise

    def __str__(self):
        return "<Repository: {}; Version: {}>".format(self.repository.name, self.number)


class RepositoryVersionContentDetails(models.Model):
    ADDED = "A"
    PRESENT = "P"
    REMOVED = "R"
    COUNT_TYPE_CHOICES = (
        (ADDED, "added"),
        (PRESENT, "present"),
        (REMOVED, "removed"),
    )

    count_type = models.CharField(max_length=1, choices=COUNT_TYPE_CHOICES)
    content_type = models.TextField()
    repository_version = models.ForeignKey(
        "RepositoryVersion", related_name="counts", on_delete=models.CASCADE
    )
    count = models.IntegerField()

    @property
    def content_href(self):
        """
        Generate URLs for the content types added, removed, or present in the RepositoryVersion.

        For each content type present in or removed from this RepositoryVersion, create the URL of
        the viewset of that variety of content along with a query parameter which filters it by
        presence in this RepositoryVersion summary.

        Args:
            obj (pulpcore.app.models.RepositoryVersion): The RepositoryVersion being serialized.
        Returns:
            dict: {<pulp_type>: <url>}
        """
        repository = self.repository_version.repository.cast()
        repository_content = RepositoryContent.objects.filter(repository=repository)
        ctype_query = Content.objects.filter(
            pulp_type=self.content_type, pk__in=repository_content.values_list("content", flat=True)
        )
        ctype_model = ctype_query.first().cast().__class__
        ctype_view = get_view_name_for_model(ctype_model, "list")
        try:
            ctype_url = reverse(ctype_view)
        except django.urls.exceptions.NoReverseMatch:
            # We've hit a content type for which there is no viewset.
            # There's nothing we can do here, except to skip it.
            return

        repository_view = get_view_name_for_model(repository.__class__, "list")

        repository_url = reverse(repository_view)
        rv_href = (
            repository_url
            + str(repository.pk)
            + "/versions/{version}/".format(version=self.repository_version.number)
        )
        if self.count_type == self.ADDED:
            partial_url_str = "{base}?repository_version_added={rv_href}"
        elif self.count_type == self.PRESENT:
            partial_url_str = "{base}?repository_version={rv_href}"
        elif self.count_type == self.REMOVED:
            partial_url_str = "{base}?repository_version_removed={rv_href}"
        full_url = partial_url_str.format(base=ctype_url, rv_href=rv_href)
        return full_url



# Generated by Django 2.2.7 on 2019-11-27 23:50

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0017_remove_task_parent'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='publication',
            name='publisher',
        ),
        migrations.DeleteModel(
            name='Publisher',
        ),
    ]



from gettext import gettext as _

from .base import PulpException


class MissingPlugin(PulpException):
    """
    Missing plugin exception.

    Exception that is raised when a requested plugin is not installed.
    """

    def __init__(self, plugin_app_label):
        """
        :param resources: keyword arguments of resource_type=resource_id
        :type resources: dict
        """
        super().__init__("PLP0002")
        self.plugin_app_label = plugin_app_label

    def __str__(self):
        msg = _("Plugin with Django app label %s is not installed.") % self.plugin_app_label
        return msg



import hashlib
import os

from django.core.files.base import ContentFile
from django.db import models
from django.db.models.signals import post_delete
from django.dispatch import receiver
from rest_framework import serializers

from pulpcore.app.models import BaseModel, fields, storage


class Upload(BaseModel):
    """
    A chunked upload. Stores chunks until used to create an artifact, etc.

    Fields:

        size (models.BigIntegerField): The size of the file in bytes.
    """

    size = models.BigIntegerField()

    def append(self, chunk, offset, sha256=None):
        """
        Append a chunk to an upload.

        Args:
            chunk (File): Binary file to append to the upload file.
            offset (int): First byte position to write chunk to.
        """
        chunk_read = chunk.read()
        current_sha256 = hashlib.sha256(chunk_read).hexdigest()
        if sha256 and sha256 != current_sha256:
            raise serializers.ValidationError("Checksum does not match chunk upload.")

        upload_chunk = UploadChunk(upload=self, offset=offset, size=len(chunk))
        filename = os.path.basename(upload_chunk.storage_path(""))
        upload_chunk.file.save(filename, ContentFile(chunk_read))


class UploadChunk(BaseModel):
    """
    A chunk for an uploaded file.

    Fields:

        file (fields.FileField): A file where the uploaded chunk is stored.
        upload (models.ForeignKey): Upload this chunk belongs to.
        offset (models.BigIntegerField): Start of the chunk in bytes.
        size (models.BigIntegerField): Size of the chunk in bytes.
    """

    def storage_path(self, name):
        """
        Callable used by FileField to determine where the uploaded file should be stored.

        Args:
            name (str): Original name of uploaded file. It is ignored by this method because the
                pulp_id is used to determine a file path instead.
        """
        return storage.get_upload_chunk_file_path(self.pulp_id)

    file = fields.FileField(null=False, upload_to=storage_path, max_length=255)
    upload = models.ForeignKey(Upload, on_delete=models.CASCADE, related_name="chunks")
    offset = models.BigIntegerField()
    size = models.BigIntegerField()


@receiver(post_delete, sender=UploadChunk)
def upload_chunk_delete(instance, **kwargs):
    instance.file.delete(save=False)



from time import sleep
from django.test import TestCase
from unittest import skipUnless

from pulpcore.cache import Cache, ConnectionError


try:
    is_redis_connected = Cache().redis.ping()
except ConnectionError:
    is_redis_connected = False


@skipUnless(is_redis_connected, "Could not connect to the Redis server")
class CacheBasicOperationsTestCase(TestCase):
    """Tests the basic APIs of the Cache object"""

    def test_01_basic_set_get(self):
        """Tests setting value, then getting it"""
        cache = Cache()
        cache.set("key", "hello")
        ret = cache.get("key")
        self.assertEqual(ret, b"hello")
        cache.set("key", "there")
        ret = cache.get("key")
        self.assertEqual(ret, b"there")

    def test_02_basic_exists(self):
        """Tests that keys already set exist"""
        cache = Cache()
        self.assertTrue(cache.exists("key"))
        self.assertFalse(cache.exists("absent"))

    def test_03_basic_delete(self):
        """Tests deleting value"""
        cache = Cache()
        self.assertTrue(cache.exists("key"))
        cache.delete("key")
        ret = cache.get("key")
        self.assertIsNone(ret)

    def test_04_basic_expires(self):
        """Tests setting values with expiration times"""
        self.skipTest("Timing is inconsistent in CI")
        cache = Cache()
        cache.set("key", "hi", expires=5)
        ret = cache.get("key")
        self.assertEqual(ret, b"hi")
        sleep(5)
        ret = cache.get("key")
        self.assertIsNone(ret)

    def test_05_group_with_base_key(self):
        """Tests grouping multiple key-values under one base-key"""
        cache = Cache()
        tuples = [
            ("key1", "hi", "base1"),
            ("key2", "friends", "base1"),
            ("key1", "hola", "base2"),
            ("key2", "amigos", "base2"),
        ]
        for key, value, base_key in tuples:
            cache.set(key, value, base_key=base_key)
        for key, value, base_key in tuples:
            self.assertEqual(value.encode(), cache.get(key, base_key=base_key))

        dict1 = {a.encode(): b.encode() for a, b, _ in tuples[:2]}
        dict2 = {a.encode(): b.encode() for a, b, _ in tuples[2:]}
        self.assertDictEqual(dict1, cache.get(None, base_key="base1"))
        self.assertDictEqual(dict2, cache.get(None, base_key="base2"))
        self.assertTrue(cache.exists(base_key="base1"))
        self.assertTrue(cache.exists(base_key="base2"))
        self.assertEqual(2, cache.exists(base_key=["base1", "base2"]))

    def test_06_delete_base_key(self):
        """Tests deleting multiple key-values under one base-key"""
        cache = Cache()
        cache.delete(base_key="base1")
        self.assertFalse(cache.exists("key1", base_key="base1"))
        self.assertFalse(cache.exists("key2", base_key="base1"))
        self.assertFalse(cache.exists(base_key="base1"))

        cache.set("key1", "hi", base_key="base1")
        self.assertTrue(cache.exists("key1", base_key="base1"))
        # multi delete
        cache.delete(base_key=["base1", "base2"])
        self.assertEqual(0, cache.exists(base_key=["base1", "base2"]))

    def test_07_clear(self):
        """Tests clearing the cache"""
        cache = Cache()
        tuples = [
            ("key", "hi", None),
            ("key1", "there", None),
            ("key", "hey", "base"),
            ("key1", "now", "base"),
        ]
        for key, value, base_key in tuples:
            cache.set(key, value, base_key=base_key)
        cache.redis.flushdb()
        for key, _, base_key in tuples:
            self.assertFalse(cache.exists(key, base_key=base_key))



from types import SimpleNamespace

#: All valid task states.
TASK_STATES = SimpleNamespace(
    WAITING="waiting",
    SKIPPED="skipped",
    RUNNING="running",
    COMPLETED="completed",
    FAILED="failed",
    CANCELED="canceled",
    CANCELING="canceling",
)

# The same as above, but in a format that choice fields can use
TASK_CHOICES = (
    (TASK_STATES.WAITING, "Waiting"),
    (TASK_STATES.SKIPPED, "Skipped"),
    (TASK_STATES.RUNNING, "Running"),
    (TASK_STATES.COMPLETED, "Completed"),
    (TASK_STATES.FAILED, "Failed"),
    (TASK_STATES.CANCELED, "Canceled"),
    (TASK_STATES.CANCELING, "Canceling"),
)

#: Tasks in a final state have finished their work.
TASK_FINAL_STATES = (
    TASK_STATES.SKIPPED,
    TASK_STATES.COMPLETED,
    TASK_STATES.FAILED,
    TASK_STATES.CANCELED,
)

#: Tasks in an incomplete state have not finished their work yet.
TASK_INCOMPLETE_STATES = (TASK_STATES.WAITING, TASK_STATES.RUNNING, TASK_STATES.CANCELING)


SYNC_MODES = SimpleNamespace(ADDITIVE="additive", MIRROR="mirror")
SYNC_CHOICES = (
    (SYNC_MODES.ADDITIVE, "Add new content from the remote repository."),
    (
        SYNC_MODES.MIRROR,
        "Add new content and remove content is no longer in the remote repository.",
    ),
)


API_ROOT = "pulp/api/v3/"

# What content-identifying checksums algorithms does pulp _know about_?
# NOTE that models.Artifact must include a field for each possible checksum; simply
# adding here won't make a new type-of checksum available. sha256 MUST be here,
# as Pulp relies on it to identify entities.
ALL_KNOWN_CONTENT_CHECKSUMS = {"md5", "sha1", "sha224", "sha256", "sha384", "sha512"}

FS_EXPORT_METHODS = SimpleNamespace(
    WRITE="write",
    HARDLINK="hardlink",
    SYMLINK="symlink",
)
FS_EXPORT_CHOICES = (
    (FS_EXPORT_METHODS.WRITE, "Export by writing"),
    (FS_EXPORT_METHODS.HARDLINK, "Export by hardlinking"),
    (FS_EXPORT_METHODS.SYMLINK, "Export by symlinking"),
)

EXPORT_BATCH_SIZE = 2000



from django.contrib import admin
from guardian.admin import GuardedModelAdmin

from pulpcore.app.models import Task, RBACContentGuard


@admin.register(Task)
class TaskAdmin(GuardedModelAdmin):
    list_display = (
        "pulp_id",
        "pulp_created",
        "pulp_last_updated",
        "state",
        "name",
        "started_at",
        "finished_at",
        "error",
        "worker",
        "parent_task",
        "task_group",
    )
    list_filter = (
        "pulp_created",
        "pulp_last_updated",
        "started_at",
        "finished_at",
    )
    raw_id_fields = ("worker",)
    search_fields = ("name",)
    readonly_fields = (
        "pulp_id",
        "pulp_created",
        "pulp_last_updated",
        "state",
        "name",
        "started_at",
        "finished_at",
        "error",
        "worker",
        "parent_task",
        "task_group",
    )


@admin.register(RBACContentGuard)
class RBACContentGuardAdmin(GuardedModelAdmin):
    list_display = (
        "name",
        "description",
    )
    list_filter = (
        "name",
        "pulp_created",
        "pulp_last_updated",
    )
    search_fields = ("name",)



"""Constants for pulpcore API tests that require the use of a plugin."""
import os
from urllib.parse import urljoin

from pulp_smash import config
from pulp_smash.pulp3.constants import (
    BASE_DISTRIBUTION_PATH,
    BASE_PATH,
    BASE_PUBLICATION_PATH,
    BASE_REMOTE_PATH,
    BASE_REPO_PATH,
    BASE_CONTENT_PATH,
)

PULP_REPOSITORY_VERSION_BASE_URL = urljoin(BASE_PATH, "repository_versions/")

PULP_PUBLICATION_BASE_URL = urljoin(BASE_PATH, "publications/")

PULP_FIXTURES_BASE_URL = config.get_config().get_fixtures_url()

PULP_CONTENT_HOST_BASE_URL = config.get_config().get_content_host_base_url()

PULP_CONTENT_BASE_URL = urljoin(PULP_CONTENT_HOST_BASE_URL, "pulp/content/")

FILE_CONTENT_NAME = "file.file"

FILE_CONTENT_PATH = urljoin(BASE_CONTENT_PATH, "file/files/")

FILE_REMOTE_PATH = urljoin(BASE_REMOTE_PATH, "file/file/")

FILE_REPO_PATH = urljoin(BASE_REPO_PATH, "file/file/")

FILE_DISTRIBUTION_PATH = urljoin(BASE_DISTRIBUTION_PATH, "file/file/")

FILE_PUBLICATION_PATH = urljoin(BASE_PUBLICATION_PATH, "file/file/")

FILE_CHUNKED_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file-chunked/")

FILE_TO_BE_CHUNKED_URL = urljoin(FILE_CHUNKED_FIXTURE_URL, "1.iso")

FILE_CHUNKED_MANIFEST_URL = urljoin(FILE_CHUNKED_FIXTURE_URL, "PULP_MANIFEST")

FILE_CHUNKED_PART_1_URL = urljoin(FILE_CHUNKED_FIXTURE_URL, "chunkaa")

FILE_CHUNKED_PART_2_URL = urljoin(FILE_CHUNKED_FIXTURE_URL, "chunkab")

FILE_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file/")
"""The URL to a file repository."""

FILE_FIXTURE_WITH_MISSING_FILES_URL = urljoin(PULP_FIXTURES_BASE_URL, "file-manifest/")
"""The URL to a file repository with missing files."""

FILE_FIXTURE_WITH_MISSING_FILES_MANIFEST_URL = urljoin(
    FILE_FIXTURE_WITH_MISSING_FILES_URL, "PULP_MANIFEST"
)
"""The URL to a file repository with missing files manifest."""

FILE_CHUNKED_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file-chunked/")
"""The URL to a file repository."""

FILE_CHUNKED_FIXTURE_MANIFEST_URL = urljoin(FILE_CHUNKED_FIXTURE_URL, "PULP_MANIFEST")
"""The URL to a file repository manifest"""

FILE_FIXTURE_MANIFEST_URL = urljoin(FILE_FIXTURE_URL, "PULP_MANIFEST")
"""The URL to a file repository manifest."""

FILE_FIXTURE_COUNT = 3
"""The number of packages available at :data:`FILE_FIXTURE_URL`."""

FILE_FIXTURE_SUMMARY = {FILE_CONTENT_NAME: FILE_FIXTURE_COUNT}
"""The desired content summary after syncing :data:`FILE_FIXTURE_URL`."""

FILE2_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file2/")
"""The URL to a file repository."""

FILE2_FIXTURE_MANIFEST_URL = urljoin(FILE2_FIXTURE_URL, "PULP_MANIFEST")
"""The URL to a file repository manifest"""

FILE_MANY_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file-many/")
"""The URL to a file repository containing many files."""

FILE_MANY_FIXTURE_MANIFEST_URL = urljoin(FILE_MANY_FIXTURE_URL, "PULP_MANIFEST")
"""The URL to a file repository manifest"""

FILE_MANY_FIXTURE_COUNT = 250
"""The number of packages available at :data:`FILE_MANY_FIXTURE_URL`."""

FILE_LARGE_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file-large/")
"""The URL to a file repository containing a large number of files."""

FILE_LARGE_URL = urljoin(FILE_LARGE_FIXTURE_URL, "1.iso")
"""The URL to a large ISO file at :data:`FILE_LARGE_FIXTURE_URL`."""

FILE_LARGE_FIXTURE_COUNT = 10
"""The number of packages available at :data:`FILE_LARGE_FIXTURE_URL`."""

FILE_LARGE_FIXTURE_MANIFEST_URL = urljoin(FILE_LARGE_FIXTURE_URL, "PULP_MANIFEST")
"""The URL to a file repository manifest."""

FILE_URL = urljoin(FILE_FIXTURE_URL, "1.iso")
"""The URL to an ISO file at :data:`FILE_FIXTURE_URL`."""

FILE2_URL = urljoin(FILE2_FIXTURE_URL, "1.iso")
"""The URL to an ISO file at :data:`FILE2_FIXTURE_URL`."""

_CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))

X509_CERTS_BASE_PATH = os.path.join(_CURRENT_DIR, "artifacts", "x509", "certificates")
X509_CA_CERT_FILE_PATH = os.path.join(X509_CERTS_BASE_PATH, "ca.pem")



from django.test import TestCase

from pulpcore.download.factory import DownloaderFactory
from pulpcore.plugin.models import Remote


class DownloaderFactoryHeadersTestCase(TestCase):
    def test_user_agent_header(self):
        remote = Remote.objects.create(url="http://example.org/", name="foo")
        factory = DownloaderFactory(remote)
        downloader = factory.build(remote.url)
        default_user_agent = DownloaderFactory.user_agent()
        self.assertEqual(downloader.session.headers["User-Agent"], default_user_agent)
        remote.delete()

    def test_custom_user_agent_header(self):
        remote = Remote.objects.create(
            url="http://example.org/", headers=[{"User-Agent": "foo"}], name="foo"
        )
        factory = DownloaderFactory(remote)
        downloader = factory.build(remote.url)
        default_user_agent = DownloaderFactory.user_agent()
        expected_user_agent = f"{default_user_agent}, foo"
        self.assertEqual(downloader.session.headers["User-Agent"], expected_user_agent)
        remote.delete()

    def test_custom_headers(self):
        remote = Remote.objects.create(
            url="http://example.org/", headers=[{"Connection": "keep-alive"}], name="foo"
        )
        factory = DownloaderFactory(remote)
        downloader = factory.build(remote.url)
        self.assertEqual(downloader.session.headers["Connection"], "keep-alive")



from django.db import IntegrityError, models, transaction
from django_lifecycle import hook, AFTER_UPDATE, BEFORE_DELETE

from .base import MasterModel, BaseModel
from .content import Artifact, Content, ContentArtifact
from .repository import Remote, Repository, RepositoryVersion
from .task import CreatedResource
from pulpcore.app.files import PulpTemporaryUploadedFile
from pulpcore.cache import Cache
from dynaconf import settings
from rest_framework.exceptions import APIException
from pulpcore.app.models import AutoAddObjPermsMixin


class PublicationQuerySet(models.QuerySet):
    """A queryset that provides publication filtering methods."""

    def with_content(self, content):
        """
        Filters publictions that contain the provided content units.

        Args:
            content (django.db.models.QuerySet): query of content

        Returns:
            django.db.models.QuerySet: Publications that contain content.
        """
        pub_artifact_q = Publication.objects.filter(
            published_artifact__content_artifact__content__pk__in=content
        )
        pass_thru_q = Publication.objects.filter(
            pass_through=True,
            repository_version__pk__in=RepositoryVersion.objects.with_content(content),
        )

        return pub_artifact_q | pass_thru_q


class Publication(MasterModel):
    """
    A publication contains metadata and artifacts associated with content
    contained within a RepositoryVersion.

    Using as a context manager is highly encouraged.  On context exit, the complete attribute
    is set True provided that an exception has not been raised.  In the event and exception
    has been raised, the publication is deleted.

    Fields:
        complete (models.BooleanField): State tracking; for internal use. Indexed.
        pass_through (models.BooleanField): Indicates that the publication is a pass-through
            to the repository version. Enabling pass-through has the same effect as creating
            a PublishedArtifact for all of the content (artifacts) in the repository.

    Relations:
        repository_version (models.ForeignKey): The RepositoryVersion used to
            create this Publication.

    Examples:
        >>> repository_version = ...
        >>>
        >>> with Publication.create(repository_version) as publication:
        >>>     for content in repository_version.content():
        >>>         for content_artifact in content.contentartifact_set.all():
        >>>             artifact = PublishedArtifact(...)
        >>>             artifact.save()
        >>>             metadata = PublishedMetadata.create_from_file(...)
        >>>             ...
        >>>
    """

    TYPE = "publication"

    objects = PublicationQuerySet.as_manager()

    complete = models.BooleanField(db_index=True, default=False)
    pass_through = models.BooleanField(default=False)

    repository_version = models.ForeignKey("RepositoryVersion", on_delete=models.CASCADE)

    @classmethod
    def create(cls, repository_version, pass_through=False):
        """
        Create a publication.

        This should be used to create a publication.  Using Publication() directly
        is highly discouraged.

        Args:
            repository_version (pulpcore.app.models.RepositoryVersion): The repository
                version to be published.
            pass_through (bool): Indicates that the publication is a pass-through
                to the repository version. Enabling pass-through has the same effect
                as creating a PublishedArtifact for all of the content (artifacts)
                in the repository.

        Returns:
            pulpcore.app.models.Publication: A created Publication in an incomplete state.

        Notes:
            Adds a Task.created_resource for the publication.
        """
        with transaction.atomic():
            publication = cls(pass_through=pass_through, repository_version=repository_version)
            publication.save()
            resource = CreatedResource(content_object=publication)
            resource.save()
            return publication

    @property
    def repository(self):
        """
        Return the associated repository

        Returns:
            pulpcore.app.models.Repository: The repository associated to this publication
        """
        return self.repository_version.repository

    def delete(self, **kwargs):
        """
        Delete the publication.

        Args:
            **kwargs (dict): Delete options.

        Notes:
            Deletes the Task.created_resource when complete is False.
        """
        with transaction.atomic():
            # invalidate cache
            if settings.CACHE_ENABLED:
                # Find any publications being served directly
                base_paths = self.distribution_set.values_list("base_path", flat=True)
                # Find any publications being served indirectly by auto-distribute feature
                # It's possible for errors to occur before any publication has been completed,
                # so we need to handle the case when no Publication exists.
                try:
                    versions = self.repository.versions.all()
                    pubs = Publication.objects.filter(
                        repository_version__in=versions, complete=True
                    )
                    publication = pubs.latest("repository_version", "pulp_created")
                    if self.pk == publication.pk:
                        base_paths |= self.repository.distributions.values_list(
                            "base_path", flat=True
                        )
                except Publication.DoesNotExist:
                    pass

                # Invalidate cache for all distributions serving this publication
                if base_paths:
                    Cache().delete(base_key=base_paths)

            CreatedResource.objects.filter(object_id=self.pk).delete()
            super().delete(**kwargs)

    def finalize_new_publication(self):
        """
        Finalize the incomplete Publication with plugin-provided code.

        This method should be overridden by plugin writers for an opportunity for plugin input. This
        method is intended to be used to validate or modify the content.

        This method does not adjust the value of complete, or save the `Publication` itself.
        Its intent is to allow the plugin writer an opportunity for plugin input before pulpcore
        marks the `Publication` as complete.
        """
        pass

    def __enter__(self):
        """
        Enter context.

        Returns:
            Publication: self
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Set the complete=True, create the publication.

        Args:
            exc_type (Type): (optional) Type of exception raised.
            exc_val (Exception): (optional) Instance of exception raised.
            exc_tb (types.TracebackType): (optional) stack trace.
        """
        if exc_val:
            # If an exception got us here, the Publication we were trying to create is
            # Bad, and we should delete the attempt. HOWEVER - some exceptions happen before we
            # even get that far. In those cases, calling delete() results in a new not-very-useful
            # exception being raised and reported to the user, rather than the actual problem.
            try:
                self.delete()
            except Exception:
                raise exc_val.with_traceback(exc_tb)
        else:
            try:
                self.finalize_new_publication()
                self.complete = True
                self.save()
            except Exception:
                self.delete()
                raise

            # invalidate cache
            if settings.CACHE_ENABLED:
                base_paths = Distribution.objects.filter(
                    repository=self.repository_version.repository
                ).values_list("base_path", flat=True)
                if base_paths:
                    Cache().delete(base_key=base_paths)


class PublishedArtifact(BaseModel):
    """
    An artifact that is part of a publication.

    Fields:
        relative_path (models.TextField): The (relative) path component of the published url.

    Relations:
        content_artifact (models.ForeignKey): The referenced content artifact.
        publication (models.ForeignKey): The publication in which the artifact is included.
    """

    relative_path = models.TextField()

    content_artifact = models.ForeignKey("ContentArtifact", on_delete=models.CASCADE)
    publication = models.ForeignKey(Publication, on_delete=models.CASCADE)

    class Meta:
        default_related_name = "published_artifact"
        unique_together = ("publication", "relative_path")


class PublishedMetadata(Content):
    """
    Metadata file that is part of a publication.

    Fields:
        relative_path (models.TextField): The (relative) path component of the published url.

    Relations:
        publication (models.ForeignKey): The publication in which the artifact is included.
    """

    TYPE = "publishedmetadata"

    relative_path = models.TextField()

    publication = models.ForeignKey(Publication, on_delete=models.CASCADE)

    @classmethod
    def create_from_file(cls, file, publication, relative_path=None):
        """
        Creates PublishedMetadata along with Artifact, ContentArtifact, and PublishedArtifact.

        Args:
            file (django.core.files.File): an open File that contains metadata
            publication (pulpcore.app.models.Publication): The publication in which the
                PublishedMetadata is included.
            relative_path (str): relative path at which the Metadata is published at. If None, the
                name of the 'file' is used.

        Returns:
            PublishedMetadata (pulpcore.app.models.PublishedMetadata):
                A saved instance of PublishedMetadata.
        """

        with transaction.atomic():
            artifact = Artifact.init_and_validate(file=PulpTemporaryUploadedFile.from_file(file))
            try:
                with transaction.atomic():
                    artifact.save()
            except IntegrityError:
                artifact = Artifact.objects.get(sha256=artifact.sha256)
            if not relative_path:
                relative_path = file.name
            content = cls(relative_path=relative_path, publication=publication)
            content.save()
            ca = ContentArtifact(relative_path=relative_path, content=content, artifact=artifact)
            ca.save()
            pa = PublishedArtifact(
                relative_path=relative_path, content_artifact=ca, publication=publication
            )
            pa.save()
        return content

    class Meta:
        default_related_name = "published_metadata"
        unique_together = ("publication", "relative_path")


class ContentGuard(MasterModel):
    """
    Defines a named content guard.

    This is meant to be subclassed by plugin authors as an opportunity to provide
    plugin-specific persistent attributes and additional validation for those attributes.
    The permit() method must be overridden to provide the web request authorization logic.

    This object is a Django model that inherits from :class: `pulpcore.app.models.ContentGuard`
    which provides the platform persistent attributes for a content-guard. Plugin authors can
    add additional persistent attributes by subclassing this class and adding Django fields.
    We defer to the Django docs on extending this model definition with additional fields.

    Fields:
        name (models.TextField): Unique guard name.
        description (models.TextField): An optional description.

    """

    name = models.TextField(db_index=True, unique=True)
    description = models.TextField(null=True)

    def permit(self, request):
        """
        Authorize the specified web request.

        Args:
            request (aiohttp.web.Request): A request for a published file.

        Raises:
            PermissionError: When not authorized.
        """
        raise NotImplementedError()

    @hook(BEFORE_DELETE)
    def invalidate_cache(self):
        if settings.CACHE_ENABLED:
            base_paths = self.distribution_set.values_list("base_path", flat=True)
            if base_paths:
                Cache().delete(base_key=base_paths)


class RBACContentGuard(ContentGuard, AutoAddObjPermsMixin):
    """
    A content guard that protects content based on RBAC permissions.
    """

    ACCESS_POLICY_VIEWSET_NAME = "contentguards/core/rbac"
    TYPE = "rbac"

    def permit(self, request):
        """
        Authorize the specified web request. Expects the request to have already been authenticated.
        """
        if not (drequest := request.get("drf_request", None)):
            raise PermissionError("Content app didn't properly authenticate this request")
        from pulpcore.app.viewsets import RBACContentGuardViewSet

        view = RBACContentGuardViewSet()
        setattr(view, "get_object", lambda: self)
        setattr(view, "action", "download")
        try:
            view.check_permissions(drequest)
        except APIException as e:
            raise PermissionError(e)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = (
            ("download_rbaccontentguard", "Can Download Content"),
            ("manage_roles_rbaccontentguard", "Can manage role assignments on RBAC content guard"),
        )


class BaseDistribution(MasterModel):
    """
    A distribution defines how a publication is distributed by the Content App.

    This abstract model can be used by plugin writers to create concrete distributions that are
    stored in separate tables from the Distributions provided by pulpcore.

    The `name` must be unique.

    The ``base_path`` must have no overlapping components. So if a Distribution with ``base_path``
    of ``a/path/foo`` existed, you could not make a second Distribution with a ``base_path`` of
    ``a/path`` or ``a`` because both are subpaths of ``a/path/foo``.

    Fields:
        name (models.TextField): The name of the distribution. Examples: "rawhide" and "stable".
        base_path (models.TextField): The base (relative) path component of the published url.

    Relations:
        content_guard (models.ForeignKey): An optional content-guard.
        remote (models.ForeignKey): A remote that the content app can use to find content not
            yet stored in Pulp.
    """

    name = models.TextField(db_index=True, unique=True)
    base_path = models.TextField(unique=True)

    content_guard = models.ForeignKey(ContentGuard, null=True, on_delete=models.SET_NULL)
    remote = models.ForeignKey(Remote, null=True, on_delete=models.SET_NULL)

    def __init__(self, *args, **kwargs):
        raise Exception(
            "BaseDistribution is no longer supported. "
            "Please use pulpcore.plugin.models.Distribution instead."
        )


class Distribution(MasterModel):
    """
    A Distribution defines how the Content App distributes a publication or repository_version.

    This master model can be used by plugin writers to create detail Distribution objects.

    The ``name`` must be unique.

    The ``base_path`` must have no overlapping components. So if a Distribution with ``base_path``
    of ``a/path/foo`` existed, you could not make a second Distribution with a ``base_path`` of
    ``a/path`` or ``a`` because both are subpaths of ``a/path/foo``.

    Subclasses are expected to use either the ``publication`` or ``repository_version`` field, but
    not both. The content app that serves content is not prepared to serve content both ways at the
    same time.

    Fields:
        name (models.TextField): The name of the distribution. Examples: "rawhide" and "stable".
        base_path (models.TextField): The base (relative) path component of the published url.

    Relations:
        content_guard (models.ForeignKey): An optional content-guard.
        publication (models.ForeignKey): Publication to be served.
        remote (models.ForeignKey): A remote that the content app can use to find content not
            yet stored in Pulp.
        repository (models.ForeignKey): The latest RepositoryVersion for this Repository will be
            served.
        repository_version (models.ForeignKey): RepositoryVersion to be served.
    """

    # If distribution serves publications, set by subclasses for proper handling in content app
    SERVE_FROM_PUBLICATION = False

    name = models.TextField(db_index=True, unique=True)
    base_path = models.TextField(unique=True)

    content_guard = models.ForeignKey(ContentGuard, null=True, on_delete=models.SET_NULL)
    publication = models.ForeignKey(Publication, null=True, on_delete=models.SET_NULL)
    remote = models.ForeignKey(Remote, null=True, on_delete=models.SET_NULL)
    repository = models.ForeignKey(
        Repository, null=True, on_delete=models.SET_NULL, related_name="distributions"
    )
    repository_version = models.ForeignKey(RepositoryVersion, null=True, on_delete=models.SET_NULL)

    def content_handler(self, path):
        """
        Handler to serve extra, non-Artifact content for this Distribution

        Args:
            path (str): The path being requested
        Returns:
            None if there is no content to be served at path. Otherwise a
            aiohttp.web_response.Response with the content.
        """
        return None

    def content_handler_list_directory(self, rel_path):
        """
        Generate the directory listing entries for content_handler

        Args:
            rel_path (str): relative path inside the distribution's base_path. For example,
            the root of the base_path is '', a subdir within the base_path is 'subdir/'.
        Returns:
            Set of strings for the extra entries in rel_path
        """
        return set()

    @hook(BEFORE_DELETE)
    @hook(
        AFTER_UPDATE,
        when_any=[
            "base_path",
            "content_guard",
            "publication",
            "remote",
            "repository",
            "repository_version",
        ],
        has_changed=True,
    )
    def invalidate_cache(self):
        """Invalidates the cache if enabled."""
        if settings.CACHE_ENABLED:
            Cache().delete(base_key=self.base_path)
            # Can also preload cache here possibly



import hashlib
import json
import logging
import os
import subprocess
import tarfile

from distutils.util import strtobool
from gettext import gettext as _
from glob import glob
from pathlib import Path
from pkg_resources import get_distribution

from django.conf import settings

from pulpcore.app.models import (
    CreatedResource,
    ExportedResource,
    Exporter,
    FilesystemExport,
    Publication,
    PulpExport,
    PulpExporter,
    RepositoryVersion,
    Task,
)
from pulpcore.app.models.content import ContentArtifact
from pulpcore.app.serializers import PulpExportSerializer
from pulpcore.app.util import get_version_from_model
from pulpcore.app.importexport import (
    export_versions,
    export_artifacts,
    export_content,
)
from pulpcore.constants import FS_EXPORT_METHODS

log = logging.getLogger(__name__)


def _export_to_file_system(path, content_artifacts, method=FS_EXPORT_METHODS.WRITE):
    """
    Export a set of ContentArtifacts to the filesystem.

    Args:
        path (str): A path to export the ContentArtifacts to
        content_artifacts (django.db.models.QuerySet): Set of ContentArtifacts to export

    Raises:
        ValidationError: When path is not in the ALLOWED_EXPORT_PATHS setting
        RuntimeError: If Artifacts are not downloaded or when trying to link non-fs files
    """
    if content_artifacts.filter(artifact=None).exists():
        RuntimeError(_("Cannot export artifacts that haven't been downloaded."))

    if (
        settings.DEFAULT_FILE_STORAGE != "pulpcore.app.models.storage.FileSystem"
        and method != FS_EXPORT_METHODS.WRITE
    ):
        raise RuntimeError(_("Only write is supported for non-filesystem storage."))

    for ca in content_artifacts.select_related("artifact").iterator():
        artifact = ca.artifact
        dest = os.path.join(path, ca.relative_path)

        os.makedirs(os.path.split(dest)[0], exist_ok=True)

        if method == FS_EXPORT_METHODS.SYMLINK:
            src = os.path.join(settings.MEDIA_ROOT, artifact.file.name)
            os.symlink(src, dest)
        elif method == FS_EXPORT_METHODS.HARDLINK:
            src = os.path.join(settings.MEDIA_ROOT, artifact.file.name)
            os.link(src, dest)
        elif method == FS_EXPORT_METHODS.WRITE:
            with open(dest, "wb") as f, artifact.file as af:
                for chunk in af.chunks(1024 * 1024):
                    f.write(chunk)

        else:
            raise RuntimeError(_("Unsupported export method '{}'.").format(method))


def fs_publication_export(exporter_pk, publication_pk):
    """
    Export a publication to the file system.

    Args:
        exporter_pk (str): FilesystemExporter pk
        publication_pk (str): Publication pk
    """
    exporter = Exporter.objects.get(pk=exporter_pk).cast()
    publication = Publication.objects.get(pk=publication_pk).cast()
    export = FilesystemExport.objects.create(
        exporter=exporter,
        params={"publication": publication_pk},
        task=Task.current(),
    )
    ExportedResource.objects.create(export=export, content_object=publication)
    CreatedResource.objects.create(content_object=export)

    log.info(
        _(
            "Exporting: file_system_exporter={exporter}, publication={publication}, path=path"
        ).format(exporter=exporter.name, publication=publication.pk, path=exporter.path)
    )

    content_artifacts = ContentArtifact.objects.filter(
        pk__in=publication.published_artifact.values_list("content_artifact__pk", flat=True)
    )

    if publication.pass_through:
        content_artifacts |= ContentArtifact.objects.filter(
            content__in=publication.repository_version.content
        )

    _export_to_file_system(exporter.path, content_artifacts)


def fs_repo_version_export(exporter_pk, repo_version_pk):
    """
    Export a repository version to the file system.

    Args:
        exporter_pk (str): FilesystemExporter pk
        repo_version_pk (str): RepositoryVersion pk
    """
    exporter = Exporter.objects.get(pk=exporter_pk).cast()
    repo_version = RepositoryVersion.objects.get(pk=repo_version_pk)
    export = FilesystemExport.objects.create(
        exporter=exporter,
        params={"repository_version": repo_version_pk},
        task=Task.current(),
    )
    ExportedResource.objects.create(export=export, content_object=repo_version)
    CreatedResource.objects.create(content_object=export)

    log.info(
        _(
            "Exporting: file_system_exporter={exporter}, repo_version={repo_version}, path=path"
        ).format(exporter=exporter.name, repo_version=repo_version.pk, path=exporter.path)
    )

    content_artifacts = ContentArtifact.objects.filter(content__in=repo_version.content)

    _export_to_file_system(exporter.path, content_artifacts)


def _get_versions_to_export(the_exporter, the_export):
    """
    Return repo-versions to be exported.

    versions is based on exporter-repositories and how the export-cmd was
    invoked.
    """
    repositories = the_exporter.repositories.all()
    # Figure out which RepositoryVersions we're going to be exporting
    # Use repo.latest unless versions= was specified
    if the_export.validated_versions is not None:
        versions = the_export.validated_versions
    else:
        versions = [r.latest_version() for r in repositories]
    return versions


def _get_starting_versions(do_incremental, the_exporter, the_export):
    # list-of-previous-versions, or None
    if do_incremental:
        if the_export.validated_start_versions:
            prev_versions = the_export.validated_start_versions
        else:
            prev_versions = [
                er.content_object
                for er in ExportedResource.objects.filter(export=the_exporter.last_export).all()
            ]
    else:
        prev_versions = None
    return prev_versions


def _get_versions_info(the_exporter):
    """
    Return plugin-version-info based on plugins are responsible for exporter-repositories.
    """
    repositories = the_exporter.repositories.all()

    # extract plugin-version-info based on the repositories we're exporting from
    vers_info = set()
    # We always need to know what version of pulpcore was in place
    vers_info.add(("pulpcore", get_distribution("pulpcore").version))
    # for each repository being exported, get the version-info for the plugin that
    # owns/controls that kind-of repository
    for r in repositories:
        vers_info.add(get_version_from_model(r.cast()))

    return vers_info


def _version_match(curr_versions, prev_versions):
    """
    Match list of repo-versions, to a prev-set, based on belonging to the same repository.

    We need this in order to be able to know how to 'diff' versions for incremental exports.

    :param curr_versions ([model.RepositoryVersion]): versions we want to export
    :param prev_versions ([model.RepositoryVersion]): last set of versions we exported (if any)
    :return: { <a_curr_version>: <matching-prev-version>, or None if no match or no prev_versions }
    """
    curr_to_repo = {v.repository: v for v in curr_versions}
    if prev_versions is None:
        return {curr_to_repo[repo]: None for repo in curr_to_repo}
    else:
        prev_to_repo = {v.repository: v for v in prev_versions}
        return {curr_to_repo[repo]: prev_to_repo[repo] for repo in curr_to_repo}


def _incremental_requested(the_export):
    """Figure out that a) an incremental is requested, and b) it's possible."""
    the_exporter = the_export.exporter
    full = the_export.params.get("full", True)
    if isinstance(full, str):
        full = bool(strtobool(full))
    starting_versions_provided = len(the_export.params.get("start_versions", [])) > 0
    last_exists = the_exporter.last_export
    return (starting_versions_provided or last_exists) and not full


def pulp_export(exporter_pk, params):
    """
    Create a PulpExport to export pulp_exporter.repositories.

    1) Spit out all Artifacts, ArtifactResource.json, and RepositoryResource.json
    2) Spit out all *resource JSONs in per-repo-version directories
    3) Compute and store the sha256 and filename of the resulting tar.gz/chunks

    Args:
        exporter_pk (str): PulpExporter
        params (dict): request data

    Raises:
        ValidationError: When path is not in the ALLOWED_EXPORT_PATHS setting,
            OR path exists and is not a directory
    """
    pulp_exporter = PulpExporter.objects.get(pk=exporter_pk)
    serializer = PulpExportSerializer(data=params, context={"exporter": pulp_exporter})
    serializer.is_valid(raise_exception=True)
    the_export = PulpExport.objects.create(exporter=pulp_exporter, params=params)
    the_export.validated_versions = serializer.validated_data.get("versions", None)
    the_export.validated_start_versions = serializer.validated_data.get("start_versions", None)
    the_export.validated_chunk_size = serializer.validated_data.get("chunk_size", None)

    try:
        the_export.task = Task.current()

        tarfile_fp = the_export.export_tarfile_path()

        path = Path(pulp_exporter.path)
        if not path.is_dir():
            path.mkdir(mode=0o775, parents=True)

        rslts = {}
        if the_export.validated_chunk_size:
            # write it into chunks
            with subprocess.Popen(
                [
                    "split",
                    "-a",
                    "4",
                    "-b",
                    str(the_export.validated_chunk_size),
                    "-d",
                    "-",
                    tarfile_fp + ".",
                ],
                stdin=subprocess.PIPE,
            ) as split_process:
                try:
                    with tarfile.open(tarfile_fp, "w|gz", fileobj=split_process.stdin) as tar:
                        _do_export(pulp_exporter, tar, the_export)
                except Exception:
                    # no matter what went wrong, we can't trust the files we (may have) created.
                    # Delete the ones we can find and pass the problem up.
                    for pathname in glob(tarfile_fp + ".*"):
                        os.remove(pathname)
                    raise
            # compute the hashes
            global_hash = hashlib.sha256()
            paths = sorted([str(Path(p)) for p in glob(tarfile_fp + ".*")])
            for a_file in paths:
                a_hash = _compute_hash(a_file, global_hash)
                rslts[a_file] = a_hash
            tarfile_hash = global_hash.hexdigest()

        else:
            # write into the file
            try:
                with tarfile.open(tarfile_fp, "w:gz") as tar:
                    _do_export(pulp_exporter, tar, the_export)
            except Exception:
                # no matter what went wrong, we can't trust the file we created.
                # Delete it if it exists and pass the problem up.
                if os.path.exists(tarfile_fp):
                    os.remove(tarfile_fp)
                raise
            # compute the hash
            tarfile_hash = _compute_hash(tarfile_fp)
            rslts[tarfile_fp] = tarfile_hash

        # store the outputfile/hash info
        the_export.output_file_info = rslts

        # write outputfile/hash info to a file 'next to' the output file(s)
        output_file_info_path = tarfile_fp.replace(".tar.gz", "-toc.json")
        with open(output_file_info_path, "w") as outfile:
            if the_export.validated_chunk_size:
                chunk_size = the_export.validated_chunk_size
            else:
                chunk_size = 0
            chunk_toc = {
                "meta": {
                    "chunk_size": chunk_size,
                    "file": os.path.basename(tarfile_fp),
                    "global_hash": tarfile_hash,
                },
                "files": {},
            }
            # Build a toc with just filenames (not the path on the exporter-machine)
            for a_path in rslts.keys():
                chunk_toc["files"][os.path.basename(a_path)] = rslts[a_path]
            json.dump(chunk_toc, outfile)

        # store toc info
        toc_hash = _compute_hash(output_file_info_path)
        the_export.output_file_info[output_file_info_path] = toc_hash
        the_export.toc_info = {"file": output_file_info_path, "sha256": toc_hash}
    finally:
        # whatever may have happened, make sure we save the export
        the_export.save()

    # If an exception was thrown, we'll never get here - which is good, because we don't want a
    # 'failed' export to be the last_export we derive the next incremental from
    # mark it as 'last'
    pulp_exporter.last_export = the_export
    # save the exporter
    pulp_exporter.save()


def _compute_hash(filename, global_hash=None):
    sha256_hash = hashlib.sha256()
    with open(filename, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
            if global_hash:
                global_hash.update(byte_block)
        return sha256_hash.hexdigest()


def _do_export(pulp_exporter, tar, the_export):
    the_export.tarfile = tar
    CreatedResource.objects.create(content_object=the_export)
    ending_versions = _get_versions_to_export(pulp_exporter, the_export)
    plugin_version_info = _get_versions_info(pulp_exporter)
    do_incremental = _incremental_requested(the_export)
    starting_versions = _get_starting_versions(do_incremental, pulp_exporter, the_export)
    vers_match = _version_match(ending_versions, starting_versions)
    # Gather up versions and artifacts
    artifacts = []
    for version in ending_versions:
        # Check version-content to make sure we're not being asked to export
        # an on_demand repo
        content_artifacts = ContentArtifact.objects.filter(content__in=version.content)
        if content_artifacts.filter(artifact=None).exists():
            RuntimeError(_("Remote artifacts cannot be exported."))

        if do_incremental:
            vers_artifacts = version.artifacts.difference(vers_match[version].artifacts).all()
        else:
            vers_artifacts = version.artifacts.all()
        artifacts.extend(vers_artifacts)
    # export plugin-version-info
    export_versions(the_export, plugin_version_info)
    # Export the top-level entities (artifacts and repositories)
    # Note: we've already handled "what about incrementals" when building the 'artifacts' list
    export_artifacts(the_export, artifacts)
    # Export the repository-version data, per-version
    for version in ending_versions:
        export_content(the_export, version)
        ExportedResource.objects.create(export=the_export, content_object=version)



"""Tests related to the tasking system."""
import unittest
from urllib.parse import urljoin

from pulp_smash import api, config, utils
from pulp_smash.pulp3.bindings import delete_orphans
from pulp_smash.pulp3.constants import TASKS_PATH
from pulp_smash.pulp3.utils import gen_remote, gen_repo, get_content_summary, sync
from requests.exceptions import HTTPError

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_FIXTURE_MANIFEST_URL,
    FILE_FIXTURE_SUMMARY,
    FILE_LARGE_FIXTURE_MANIFEST_URL,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
)
from pulpcore.tests.functional.api.using_plugin.utils import gen_file_remote
from pulpcore.tests.functional.api.using_plugin.utils import set_up_module as setUpModule  # noqa


class MultiResourceLockingTestCase(unittest.TestCase):
    """Verify multi-resourcing locking.

    This test targets the following issues:

    * `Pulp #3186 <https://pulp.plan.io/issues/3186>`_
    * `Pulp Smash #879 <https://github.com/pulp/pulp-smash/issues/879>`_
    """

    def test_all(self):
        """Verify multi-resourcing locking.

        Do the following:

        1. Create a repository, and a remote.
        2. Update the remote to point to a different url.
        3. Immediately run a sync. The sync should fire after the update and
           sync from the second url.
        4. Assert that remote url was updated.
        5. Assert that the number of units present in the repository is
           according to the updated url.
        """
        cfg = config.get_config()
        client = api.Client(cfg, api.json_handler)

        repo = client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["pulp_href"])

        body = gen_file_remote(url=FILE_LARGE_FIXTURE_MANIFEST_URL)
        remote = client.post(FILE_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote["pulp_href"])

        url = {"url": FILE_FIXTURE_MANIFEST_URL}
        client.patch(remote["pulp_href"], url)

        sync(cfg, remote, repo)

        repo = client.get(repo["pulp_href"])
        remote = client.get(remote["pulp_href"])
        self.assertEqual(remote["url"], url["url"])
        self.assertDictEqual(get_content_summary(repo), FILE_FIXTURE_SUMMARY)


class CancelTaskTestCase(unittest.TestCase):
    """Test to cancel a task in different states.

    This test targets the following issue:

    * `Pulp #3527 <https://pulp.plan.io/issues/3527>`_
    * `Pulp #3634 <https://pulp.plan.io/issues/3634>`_
    * `Pulp Smash #976 <https://github.com/pulp/pulp-smash/issues/976>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.page_handler)

    def test_cancel_running_task(self):
        """Cancel a running task."""
        task = self.create_long_task()
        response = self.cancel_task(task)
        self.assertIsNone(response["finished_at"], response)
        self.assertIn(response["state"], ["canceling", "canceled"], response)

    def test_cancel_nonexistent_task(self):
        """Cancel a nonexistent task."""
        task_href = urljoin(TASKS_PATH, utils.uuid4() + "/")
        with self.assertRaises(HTTPError) as ctx:
            self.client.patch(task_href, json={"state": "canceled"})
        for key in ("not", "found"):
            self.assertIn(
                key, ctx.exception.response.json()["detail"].lower(), ctx.exception.response
            )

    def test_cancel_finished_task(self):
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        repo["name"] = utils.uuid4()
        task_href = self.client.patch(repo["pulp_href"], json=repo)
        with self.assertRaises(HTTPError) as ctx:
            self.cancel_task(task_href)
        self.assertEqual(ctx.exception.response.status_code, 409)
        self.assertEqual(ctx.exception.response.json()["state"], "completed")

    def test_delete_running_task(self):
        """Delete a running task."""
        task = self.create_long_task()
        with self.assertRaises(HTTPError):
            self.client.delete(task["task"])

    def create_long_task(self):
        """Create a long task. Sync a repository with large files."""
        # to force the download of files.
        delete_orphans()

        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        body = gen_remote(url=FILE_LARGE_FIXTURE_MANIFEST_URL)
        remote = self.client.post(FILE_REMOTE_PATH, body)
        self.addCleanup(self.client.delete, remote["pulp_href"])

        # use code_handler to avoid wait to the task to be completed.
        return (
            self.client.using_handler(api.code_handler)
            .post(urljoin(repo["pulp_href"], "sync/"), {"remote": remote["pulp_href"]})
            .json()
        )

    def cancel_task(self, task):
        """Cancel a task."""
        return self.client.patch(task["task"], json={"state": "canceled"})



from pulpcore.app.tasks import base, repository, upload  # noqa

from .base import (  # noqa
    general_create,
    general_create_from_temp_file,
    general_delete,
    general_multi_delete,
    general_update,
)

from .export import fs_publication_export, fs_repo_version_export  # noqa

from .importer import pulp_import  # noqa

from .orphan import orphan_cleanup  # noqa

from .purge import purge  # noqa

from .reclaim_space import reclaim_space  # noqa

from .repository import repair_all_artifacts  # noqa



"""Tests that perform actions over artifacts."""
import hashlib
import itertools
import os
import unittest

from pulp_smash import api, cli, config, utils
from pulp_smash.exceptions import CalledProcessError
from pulp_smash.pulp3.bindings import delete_orphans
from pulp_smash.pulp3.constants import ARTIFACTS_PATH
from requests.exceptions import HTTPError

# This import is an exception, we use a file url but we are not actually using
# any plugin
from pulpcore.tests.functional.api.using_plugin.constants import FILE_URL
from pulpcore.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class ArtifactTestCase(unittest.TestCase):
    """Create an artifact by uploading a file.

    This test targets the following issues:

    * `Pulp #2843 <https://pulp.plan.io/issues/2843>`_
    * `Pulp Smash #726 <https://github.com/pulp/pulp-smash/issues/726>`_
    """

    @classmethod
    def setUpClass(cls):
        """Delete orphans and create class-wide variables."""
        cfg = config.get_config()
        delete_orphans()
        cls.client = api.Client(cfg, api.json_handler)
        cls.file = {"file": utils.http_get(FILE_URL)}
        cls.file_sha256 = hashlib.sha256(cls.file["file"]).hexdigest()
        cls.file_size = len(cls.file["file"])

    def test_upload_valid_attrs(self):
        """Upload a file, and provide valid attributes.

        For each possible combination of ``sha256`` and ``size`` (including
        neither), do the following:

        1. Upload a file with the chosen combination of attributes.
        2. Verify that an artifact has been created, and that it has valid
           attributes.
        3. Delete the artifact, and verify that its attributes are
           inaccessible.
        """
        file_attrs = {"sha256": self.file_sha256, "size": self.file_size}
        for i in range(len(file_attrs) + 1):
            for keys in itertools.combinations(file_attrs, i):
                data = {key: file_attrs[key] for key in keys}
                with self.subTest(data=data):
                    self._do_upload_valid_attrs(data, self.file)

    def test_upload_empty_file(self):
        """Upload an empty file.

        For each possible combination of ``sha256`` and ``size`` (including
        neither), do the following:

        1. Upload a file with the chosen combination of attributes.
        2. Verify that an artifact has been created, and that it has valid
           attributes.
        3. Delete the artifact, and verify that its attributes are
           inaccessible.
        """
        empty_file = b""
        file_attrs = {"sha256": hashlib.sha256(empty_file).hexdigest(), "size": 0}
        for i in range(len(file_attrs) + 1):
            for keys in itertools.combinations(file_attrs, i):
                data = {key: file_attrs[key] for key in keys}
                with self.subTest(data=data):
                    self._do_upload_valid_attrs(data, files={"file": empty_file})

    def _do_upload_valid_attrs(self, data, files):
        """Upload a file with the given attributes."""
        artifact = self.client.post(ARTIFACTS_PATH, data=data, files=files)
        # assumes ALLOWED_CONTENT_CHECKSUMS does NOT contain "md5"
        self.assertTrue(artifact["md5"] is None, "MD5 {}".format(artifact["md5"]))
        self.addCleanup(self.client.delete, artifact["pulp_href"])
        read_artifact = self.client.get(artifact["pulp_href"])
        # assumes ALLOWED_CONTENT_CHECKSUMS does NOT contain "md5"
        self.assertTrue(read_artifact["md5"] is None)
        for key, val in artifact.items():
            with self.subTest(key=key):
                self.assertEqual(read_artifact[key], val)
        self.doCleanups()
        with self.assertRaises(HTTPError):
            self.client.get(artifact["pulp_href"])

    def test_upload_invalid_attrs(self):
        """Upload a file, and provide invalid attributes.

        For each possible combination of ``sha256`` and ``size`` (except for
        neither), do the following:

        1. Upload a file with the chosen combination of attributes. Verify that
           an error is returned.
        2. Verify that no artifacts exist in Pulp whose attributes match the
           file that was unsuccessfully uploaded.
        """
        file_attrs = {"sha256": utils.uuid4(), "size": self.file_size + 1}
        for i in range(1, len(file_attrs) + 1):
            for keys in itertools.combinations(file_attrs, i):
                data = {key: file_attrs[key] for key in keys}
                with self.subTest(data=data):
                    self._do_upload_invalid_attrs(data)

    def _do_upload_invalid_attrs(self, data):
        """Upload a file with the given attributes."""
        with self.assertRaises(HTTPError):
            self.client.post(ARTIFACTS_PATH, data=data, files=self.file)
        for artifact in self.client.get(ARTIFACTS_PATH)["results"]:
            self.assertNotEqual(artifact["sha256"], self.file_sha256)

    def test_upload_md5(self):
        """Attempt to upload a file using an MD5 checksum.

        Assumes ALLOWED_CONTENT_CHECKSUMS does NOT contain ``md5``
        """
        file_attrs = {"md5": utils.uuid4(), "size": self.file_size}
        with self.assertRaises(HTTPError):
            self.client.post(ARTIFACTS_PATH, data=file_attrs, files=self.file)

    def test_upload_mixed_attrs(self):
        """Upload a file, and provide both valid and invalid attributes.

        Do the following:

        1. Upload a file and provide both an ``sha256`` and a ``size``. Let one
           be valid, and the other be valid. Verify that an error is returned.
        2. Verify that no artifacts exist in Pulp whose attributes match the
           file that was unsuccessfully uploaded.
        """
        invalid_data = (
            {"sha256": self.file_sha256, "size": self.file_size + 1},
            {"sha256": utils.uuid4(), "size": self.file_size},
        )
        for data in invalid_data:
            with self.subTest(data=data):
                self._do_upload_invalid_attrs(data)


class ArtifactsDeleteFileSystemTestCase(unittest.TestCase):
    """Delete an artifact, it is removed from the filesystem.

    This test targets the following issues:

    * `Pulp #3508 <https://pulp.plan.io/issues/3508>`_
    * `Pulp Smash #908 <https://github.com/pulp/pulp-smash/issues/908>`_
    """

    def test_all(self):
        """Delete an artifact, it is removed from the filesystem.

        Do the following:

        1. Create an artifact, and verify it is present on the filesystem.
        2. Delete the artifact, and verify it is absent on the filesystem.
        """
        cfg = config.get_config()
        cli_client = cli.Client(cfg)
        storage = utils.get_pulp_setting(cli_client, "DEFAULT_FILE_STORAGE")
        if storage != "pulpcore.app.models.storage.FileSystem":
            self.skipTest("this test only works for filesystem storage")
        media_root = utils.get_pulp_setting(cli_client, "MEDIA_ROOT")

        api_client = api.Client(cfg, api.json_handler)

        # create
        files = {"file": utils.http_get(FILE_URL)}
        artifact = api_client.post(ARTIFACTS_PATH, files=files)
        self.addCleanup(api_client.delete, artifact["pulp_href"])
        cmd = ("ls", os.path.join(media_root, artifact["file"]))
        cli_client.run(cmd, sudo=True)

        # delete
        self.doCleanups()
        with self.assertRaises(CalledProcessError):
            cli_client.run(cmd, sudo=True)



"""Tests related to content upload."""
import hashlib
import unittest

from random import shuffle
from requests import HTTPError
from urllib.parse import urljoin

from pulp_smash import api, cli, config
from pulp_smash.pulp3.constants import UPLOAD_PATH
from pulp_smash.utils import http_get

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_CHUNKED_PART_1_URL,
    FILE_CHUNKED_PART_2_URL,
    FILE_TO_BE_CHUNKED_URL,
)
from pulpcore.tests.functional.api.using_plugin.utils import (  # noqa:F401
    set_up_module as setUpModule,
)


class ChunkedUploadTestCase(unittest.TestCase):
    """Test upload of files in chunks.

    This test targets the following issues:

    * `Pulp #4197 <https://pulp.plan.io/issues/4197>`_
    * `Pulp #4982 <https://pulp.plan.io/issues/4982>`_
    * `Pulp #5092 <https://pulp.plan.io/issues/5092>`_
    * `Pulp #5150 <https://pulp.plan.io/issues/5150>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.cli_client = cli.Client(cls.cfg)
        cls.client = api.Client(cls.cfg)

        cls.file = http_get(FILE_TO_BE_CHUNKED_URL)
        cls.file_sha256 = hashlib.sha256(cls.file).hexdigest()
        cls.size_file = len(cls.file)

        first_chunk = http_get(FILE_CHUNKED_PART_1_URL)
        header_first_chunk = {
            "Content-Range": "bytes 0-{}/{}".format(len(first_chunk) - 1, cls.size_file)
        }

        second_chunk = http_get(FILE_CHUNKED_PART_2_URL)
        header_second_chunk = {
            "Content-Range": "bytes {}-{}/{}".format(
                len(first_chunk), cls.size_file - 1, cls.size_file
            )
        }

        cls.chunked_data = [
            [first_chunk, header_first_chunk],
            [second_chunk, header_second_chunk],
        ]
        shuffle(cls.chunked_data)

    def test_create_artifact_without_checksum(self):
        """Test creation of artifact using upload of files in chunks."""

        _, artifact = self.upload_chunks()

        self.addCleanup(self.client.delete, artifact["pulp_href"])

        self.assertEqual(artifact["sha256"], self.file_sha256, artifact)

    def test_create_artifact_passing_checksum(self):
        """Test creation of artifact using upload of files in chunks passing checksum."""
        upload_request = self.client.post(UPLOAD_PATH, {"size": self.size_file})

        for data in self.chunked_data:
            self.client.put(
                upload_request["pulp_href"],
                data={"sha256": hashlib.sha256(data[0]).hexdigest()},
                files={"file": data[0]},
                headers=data[1],
            )

        artifact_request = self.client.post(
            urljoin(upload_request["pulp_href"], "commit/"), data={"sha256": self.file_sha256}
        )

        self.addCleanup(self.client.delete, artifact_request["pulp_href"])

        self.assertEqual(artifact_request["sha256"], self.file_sha256, artifact_request)

    def test_upload_chunk_wrong_checksum(self):
        """Test creation of artifact using upload of files in chunks passing wrong checksum."""
        upload_request = self.client.post(UPLOAD_PATH, {"size": self.size_file})

        for data in self.chunked_data:
            response = self.client.using_handler(api.echo_handler).put(
                upload_request["pulp_href"],
                data={"sha256": "WRONG CHECKSUM"},
                files={"file": data[0]},
                headers=data[1],
            )
            with self.subTest(response=response):
                self.assertEqual(response.status_code, 400, response)

        self.addCleanup(self.client.delete, upload_request["pulp_href"])

    def test_upload_response(self):
        """Test upload responses when creating an upload and uploading chunks."""
        upload_request = self.client.post(UPLOAD_PATH, {"size": self.size_file})

        expected_keys = ["pulp_href", "pulp_created", "size"]

        self.assertEqual([*upload_request], expected_keys, upload_request)

        for data in self.chunked_data:
            response = self.client.put(
                upload_request["pulp_href"], files={"file": data[0]}, headers=data[1]
            )

            with self.subTest(response=response):
                self.assertEqual([*response], expected_keys, response)

        response = self.client.get(upload_request["pulp_href"])

        expected_keys.append("chunks")

        self.assertEqual([*response], expected_keys, response)

        expected_chunks = [
            {"offset": 0, "size": 6291456},
            {"offset": 6291456, "size": 4194304},
        ]

        sorted_chunks_response = sorted(response["chunks"], key=lambda i: i["offset"])
        self.assertEqual(sorted_chunks_response, expected_chunks, response)
        self.addCleanup(self.client.delete, response["pulp_href"])

    def test_delete_upload(self):
        """Check whether uploads are being correctly deleted after committing."""
        upload, artifact = self.upload_chunks()

        with self.assertRaises(HTTPError):
            self.client.get(upload["pulp_href"])

        self.addCleanup(self.client.delete, artifact["pulp_href"])

    def upload_chunks(self):
        """Upload file in chunks."""
        upload_request = self.client.post(UPLOAD_PATH, {"size": self.size_file})

        for data in self.chunked_data:
            self.client.put(upload_request["pulp_href"], files={"file": data[0]}, headers=data[1])

        artifact_request = self.client.post(
            urljoin(upload_request["pulp_href"], "commit/"), data={"sha256": self.file_sha256}
        )
        return upload_request, artifact_request



from gettext import gettext as _

from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.models import Permission

from pulpcore.app.models.role import GroupRole


class ObjectRolePermissionBackend(BaseBackend):
    def has_perm(self, user_obj, perm, obj=None):
        if not user_obj.is_authenticated:
            return False

        try:
            app_label, codename = perm.split(".", maxsplit=1)
            permission = Permission.objects.get(
                content_type__app_label=app_label, codename=codename
            )
        except Permission.DoesNotExist:
            # Cannot have a permission that does not exist.
            return False

        if obj is None:
            # Check for global roles
            if user_obj.object_roles.filter(object_id=None, role__permissions=permission).exists():
                return True
            if GroupRole.objects.filter(
                group__in=user_obj.groups.all(), object_id=None, role__permissions=permission
            ).exists():
                return True

        else:
            obj_type = ContentType.objects.get_for_model(obj, for_concrete_model=False)
            if permission.content_type == obj_type:
                # Check object specific roles
                if user_obj.object_roles.filter(
                    object_id=obj.pk, role__permissions=permission
                ).exists():
                    return True
                if GroupRole.objects.filter(
                    group__in=user_obj.groups.all(), object_id=obj.pk, role__permissions=permission
                ).exists():
                    return True
            else:
                raise RuntimeError(
                    _("Permission {} is not suitable for objects of class {}.").format(
                        perm, obj.__class__
                    )
                )
        return False

    def get_all_permissions(self, user_obj, obj=None):
        if obj is None:
            result = (
                user_obj.object_roles.filter(object_id=None)
                .values("role__permissions__content_type__app_label", "role__permissions__codename")
                .distinct()
            )
            group_result = (
                GroupRole.objects.filter(group__in=user_obj.groups.all(), object_id=None)
                .values("role__permissions__content_type__app_label", "role__permissions__codename")
                .distinct()
            )
            return [
                item["role__permissions__content_type__app_label"]
                + "."
                + item["role__permissions__codename"]
                for item in result
            ] + [
                item["role__permissions__content_type__app_label"]
                + "."
                + item["role__permissions__codename"]
                for item in group_result
            ]

        else:
            obj_type = ContentType.objects.get_for_model(obj, for_concrete_model=False)
            # Maybe there is a way to reformulate this as a single query on Permissions
            result = (
                user_obj.object_roles.filter(role__permissions__content_type=obj_type)
                .filter(
                    content_type=obj_type,
                    object_id=obj.pk,
                )
                .values("role__permissions__codename")
                .distinct()
            )
            group_result = (
                GroupRole.objects.filter(
                    group__in=user_obj.groups.all(), role__permissions__content_type=obj_type
                )
                .filter(
                    content_type=obj_type,
                    object_id=obj.pk,
                )
                .values("role__permissions__codename")
                .distinct()
            )
        return [item["role__permissions__codename"] for item in result] + [
            item["role__permissions__codename"] for item in group_result
        ]



# Generated by Django 2.2.14 on 2020-07-22 19:50

from django.db import migrations, models
import pulpcore.app.models.content
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0036_unprotect_last_export'),
    ]

    operations = [
        migrations.CreateModel(
            name='PulpTemporaryFile',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('file', models.FileField(max_length=255, upload_to=pulpcore.app.models.content.PulpTemporaryFile.storage_path)),
            ],
            options={
                'abstract': False,
            },
            bases=(pulpcore.app.models.content.HandleTempFilesMixin, models.Model),
        ),
    ]



from datetime import datetime, timedelta, timezone
from gettext import gettext as _

from rest_framework import serializers

from pulpcore.app.serializers import ValidateFieldsMixin  # noqa
from pulpcore.constants import TASK_FINAL_STATES


class PurgeSerializer(serializers.Serializer, ValidateFieldsMixin):
    finished_before = serializers.DateTimeField(
        help_text=_(
            "Purge tasks completed earlier than this timestamp. Format '%Y-%m-%d[T%H:%M:%S]'"
        ),
        default=(datetime.now(timezone.utc) - timedelta(days=30)).strftime("%Y-%m-%d"),
    )
    states = serializers.MultipleChoiceField(
        choices=TASK_FINAL_STATES,
        default=["completed"],
        help_text=_("List of task-states to be purged. Only 'final' states are allowed."),
    )



"""
Django models related to the Tasking system
"""
import logging
import traceback
import os
from datetime import timedelta
from gettext import gettext as _

from django.contrib.postgres.fields import ArrayField
from django.db import connection, models
from django.utils import timezone
from django.conf import settings

from pulpcore.app.models import (
    AutoAddObjPermsMixin,
    AutoDeleteObjPermsMixin,
    BaseModel,
    GenericRelationModel,
)
from pulpcore.constants import TASK_CHOICES, TASK_FINAL_STATES, TASK_STATES
from pulpcore.exceptions import AdvisoryLockError, exception_to_dict
from pulpcore.tasking.constants import TASKING_CONSTANTS

_logger = logging.getLogger(__name__)


class WorkerManager(models.Manager):
    def online_workers(self):
        """
        Returns a queryset of workers meeting the criteria to be considered 'online'

        To be considered 'online', a worker must have a recent heartbeat timestamp. "Recent" is
        defined here as "within the pulp process timeout interval".

        Returns:
            :class:`django.db.models.query.QuerySet`:  A query set of the Worker objects which
                are considered by Pulp to be 'online'.
        """
        now = timezone.now()
        age_threshold = now - timedelta(seconds=settings.WORKER_TTL)

        return self.filter(last_heartbeat__gte=age_threshold)

    def missing_workers(self, age=timedelta(seconds=settings.WORKER_TTL)):
        """
        Returns a queryset of workers meeting the criteria to be considered 'missing'

        To be considered missing, a worker must have a stale timestamp.  By default, stale is
        defined here as longer than the ``settings.WORKER_TTL``, or you can specify age as a
        timedelta.

        Args:
            age (datetime.timedelta): Workers who have heartbeats older than this time interval are
                considered missing.

        Returns:
            :class:`django.db.models.query.QuerySet`:  A query set of the Worker objects which
                are considered by Pulp to be 'missing'.
        """
        age_threshold = timezone.now() - age
        return self.filter(last_heartbeat__lt=age_threshold)

    def resource_managers(self):
        """
        Returns a queryset of resource managers.

        Resource managers are identified by their name. Note that some of these may be offline.

        Returns:
            :class:`django.db.models.query.QuerySet`:  A query set of the Worker objects which
                which match the resource manager name.
        """
        return self.filter(name=TASKING_CONSTANTS.RESOURCE_MANAGER_WORKER_NAME)


class Worker(BaseModel):
    """
    Represents a worker

    Fields:

        name (models.TextField): The name of the worker, in the format "worker_type@hostname"
        last_heartbeat (models.DateTimeField): A timestamp of this worker's last heartbeat
    """

    objects = WorkerManager()

    name = models.TextField(db_index=True, unique=True)
    last_heartbeat = models.DateTimeField(auto_now=True)

    @property
    def online(self):
        """
        Whether a worker can be considered 'online'

        To be considered 'online', a worker must have a recent heartbeat timestamp. "Recent" is
        defined here as "within the pulp process timeout interval".

        Returns:
            bool: True if the worker is considered online, otherwise False
        """
        now = timezone.now()
        age_threshold = now - timedelta(seconds=settings.WORKER_TTL)

        return self.last_heartbeat >= age_threshold

    @property
    def missing(self):
        """
        Whether a worker can be considered 'missing'

        To be considered 'missing', a worker must have a stale timestamp meaning that it was not
        shutdown 'cleanly' and may have died.  Stale is defined here as "beyond the pulp process
        timeout interval".

        Returns:
            bool: True if the worker is considered missing, otherwise False
        """
        now = timezone.now()
        age_threshold = now - timedelta(seconds=settings.WORKER_TTL)

        return self.last_heartbeat < age_threshold

    def save_heartbeat(self):
        """
        Update the last_heartbeat field to now and save it.

        Only the last_heartbeat field will be saved. No other changes will be saved.

        Raises:
            ValueError: When the model instance has never been saved before. This method can
                only update an existing database record.
        """
        self.save(update_fields=["last_heartbeat"])


def _uuid_to_advisory_lock(value):
    return ((value >> 64) ^ value) & 0x7FFFFFFFFFFFFFFF


class Task(BaseModel, AutoDeleteObjPermsMixin, AutoAddObjPermsMixin):
    """
    Represents a task

    Fields:

        state (models.TextField): The state of the task
        name (models.TextField): The name of the task
        logging_cid (models.CharField): The logging CID associated with the task
        started_at (models.DateTimeField): The time the task started executing
        finished_at (models.DateTimeField): The time the task finished executing
        error (models.JSONField): Fatal errors generated by the task
        args (models.JSONField): The JSON serialized arguments for the task
        kwargs (models.JSONField): The JSON serialized keyword arguments for
            the task
        reserved_resources_record (django.contrib.postgres.fields.ArrayField): The reserved
            resources required for the task.

    Relations:

        parent (models.ForeignKey): Task that spawned this task (if any)
        worker (models.ForeignKey): The worker that this task is in
    """

    state = models.TextField(choices=TASK_CHOICES)
    name = models.TextField()
    logging_cid = models.CharField(max_length=256, db_index=True, default="")

    started_at = models.DateTimeField(null=True)
    finished_at = models.DateTimeField(null=True)

    error = models.JSONField(null=True)

    args = models.JSONField(null=True)
    kwargs = models.JSONField(null=True)

    worker = models.ForeignKey("Worker", null=True, related_name="tasks", on_delete=models.SET_NULL)

    parent_task = models.ForeignKey(
        "Task", null=True, related_name="child_tasks", on_delete=models.SET_NULL
    )
    task_group = models.ForeignKey(
        "TaskGroup", null=True, related_name="tasks", on_delete=models.SET_NULL
    )
    reserved_resources_record = ArrayField(models.CharField(max_length=256), null=True)

    ACCESS_POLICY_VIEWSET_NAME = "tasks"

    def __str__(self):
        return "Task: {name} [{state}]".format(name=self.name, state=self.state)

    def __enter__(self):
        self.lock = _uuid_to_advisory_lock(self.pk.int)
        with connection.cursor() as cursor:
            cursor.execute("SELECT pg_try_advisory_lock(%s);", [self.lock])
            acquired = cursor.fetchone()[0]
        if not acquired:
            raise AdvisoryLockError("Could not acquire lock.")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        with connection.cursor() as cursor:
            cursor.execute("SELECT pg_advisory_unlock(%s);", [self.lock])
            released = cursor.fetchone()[0]
        if not released:
            raise RuntimeError("Lock not held.")

    @staticmethod
    def current():
        """
        Returns:
            pulpcore.app.models.Task: The current task.
        """
        try:
            task_id = os.environ["PULP_TASK_ID"]
        except KeyError:
            task = None
        else:
            task = Task.objects.get(pk=task_id)
        return task

    def set_running(self):
        """
        Set this Task to the running state, save it, and log output in warning cases.

        This updates the :attr:`started_at` and sets the :attr:`state` to :attr:`RUNNING`.
        """
        rows = Task.objects.filter(pk=self.pk, state=TASK_STATES.WAITING).update(
            state=TASK_STATES.RUNNING, started_at=timezone.now()
        )
        if rows != 1:
            _logger.warning(_("Task __call__() occurred but Task %s is not at WAITING") % self.pk)
        self.refresh_from_db()

    def set_completed(self):
        """
        Set this Task to the completed state, save it, and log output in warning cases.

        This updates the :attr:`finished_at` and sets the :attr:`state` to :attr:`COMPLETED`.
        """
        # Only set the state to finished if it's not already in a complete state. This is
        # important for when the task has been canceled, so we don't move the task from canceled
        # to finished.
        rows = (
            Task.objects.filter(pk=self.pk)
            .exclude(state__in=TASK_FINAL_STATES)
            .update(state=TASK_STATES.COMPLETED, finished_at=timezone.now())
        )
        if rows != 1:
            msg = _("Task set_completed() occurred but Task %s is already in final state")
            _logger.warning(msg % self.pk)
        self.refresh_from_db()

    def set_failed(self, exc, tb):
        """
        Set this Task to the failed state and save it.

        This updates the :attr:`finished_at` attribute, sets the :attr:`state` to
        :attr:`FAILED`, and sets the :attr:`error` attribute.

        Args:
            exc (Exception): The exception raised by the task.
            tb (traceback): Traceback instance for the current exception.
        """
        tb_str = "".join(traceback.format_tb(tb))
        rows = (
            Task.objects.filter(pk=self.pk)
            .exclude(state__in=TASK_FINAL_STATES)
            .update(
                state=TASK_STATES.FAILED,
                finished_at=timezone.now(),
                error=exception_to_dict(exc, tb_str),
            )
        )
        if rows != 1:
            raise RuntimeError("Attempt to set a finished task to failed.")
        self.refresh_from_db()

    class Meta:
        indexes = [models.Index(fields=["pulp_created"])]
        permissions = [
            ("manage_roles_task", "Can manage role assignments on task"),
        ]


class TaskGroup(BaseModel):
    description = models.TextField()
    all_tasks_dispatched = models.BooleanField(default=False)

    @staticmethod
    def current():
        """
        Returns:
            pulpcore.app.models.TaskGroup: The task group the current task is being executed and
            belongs to.
        """
        try:
            task_group = Task.current().task_group
        except AttributeError:
            task_group = None
        return task_group

    def finish(self):
        """
        Finalize the task group.

        Set 'all_tasks_dispatched' to True so that API users can know that there are no
        tasks in the group yet to be created.
        """
        self.all_tasks_dispatched = True
        self.save()


class CreatedResource(GenericRelationModel):
    """
    Resources created by the task.

    Relations:
        task (models.ForeignKey): The task that created the resource.
    """

    task = models.ForeignKey(
        Task, related_name="created_resources", default=Task.current, on_delete=models.CASCADE
    )
